/* ---------------------------------------------------------- */
/* Simple  uniform  traffic  pattern,  standard  messages and */
/* packets,   message   length   and    inter-arrival    time */
/* exponentially distributed                                  */
/* ---------------------------------------------------------- */

void    initUTraffic (double mit, double mle) {

#ifdef	FIXED_MESSAGE_LENGTH
	UTPattern = create UTraffic (MIT_exp+MLE_unf, mit, mle, mle);
#else
	UTPattern = create UTraffic (MIT_exp+MLE_exp, mit, mle);
#endif
	UTPattern->addSender ();
	UTPattern->addReceiver ();
	UTPattern->printDef ("Uniform traffic pattern parameters:");
		
#ifdef	LOCAL_MEASURES
	for (int i = 0; i < NStations; i++) {
	    Node *s;
	    TheStation = s = (Node*) idToStation (i);
	    s->UAMDel = create RVariable;
            s->UAPAcc = create RVariable;
	}
#endif
}

#ifdef	LOCAL_MEASURES
#ifdef	LOCAL_MESSAGE_ACCESS_TIME
void UTraffic::pfmMTR (Packet *p) {
#else
void UTraffic::pfmMRC (Packet *p) {
#endif

	double	d;

	d = Itu * (double) (Time - p->QTime);
	((Node*) idToStation (p->Sender)) -> UAMDel -> update (d);
};
	
void UTraffic::pfmPTR (Packet *p) {

	double	d;

	d = Itu * (double) (Time - p->TTime);
	((Node*) idToStation (p->Sender)) -> UAPAcc -> update (d);
};

UTraffic::exposure {
	Node *s;

	Traffic::expose;

	onpaper {

	  exmode 8:

	    print ("Local packet access time:\n\n");
	    for (int i = 0; i < NStations; i++) {
		s = (Node*)(idToStation (i));
		if (i == 0)
			s->UAPAcc->printACnt ();
		else
			s->UAPAcc->printSCnt ();
	    }
	    print ("\n\n");

	  exmode 9:

#ifdef	LOCAL_MESSAGE_ACCESS_TIME
	    print ("Local message access time:\n\n");
#else
	    print ("Local message delay:\n\n");
#endif
	    for (i = 0; i < NStations; i++) {
		s = (Node*)(idToStation (i));
		if (i == 0)
			s->UAMDel->printACnt ();
		else
			s->UAMDel->printSCnt ();
	    }
	    print ("\n\n");

	  exmode 10:

	    print ("UT Message queues at the end of simulation:\n\n");
	    print ("Station    Messages        Bits\n");
	    for (i = 0; i < NStations; i++) {
	        long bcnt = 0L;
	        int  mcnt = 0;
	        Message **mq = idToStation (i)->MQHead, *m;

	        print (i, 7);
	        if (mq != NULL) {
	    	    for (m = mq [UTPattern->getId ()]; m != NULL; mcnt++,
			bcnt += m->Length, m = m->next);
	        }
	        print (mcnt, 12); print (bcnt, 12); print ("\n");
	    }
	    print ("\n");
	}

	onscreen {

	  exmode 8:

	    // Local packet access time
	    {
		double min, max, mmnts [2]; long ns;
	        for (int i = 0; i < NStations; i++) {
		    s = (Node*)(idToStation (i));
		    s->UAPAcc->calculate (min, max, mmnts, ns);
		    display (i);
		    display (mmnts [0]);
		    display (min);
		    display (max);
		    display (ns);
	        }
	    }

	  exmode 9:

	    // Local message delay
	    {
		double min, max, mmnts [2]; long ns;
	        for (int i = 0; i < NStations; i++) {
		    s = (Node*)(idToStation (i));
		    s->UAMDel->calculate (min, max, mmnts, ns);
		    display (i);
		    display (mmnts [0]);
		    display (min);
		    display (max);
		    display (ns);
	        }
	    }

	  exmode 10:

	    // Pending messages
	    for (int i = 0; i < NStations; i++) {
	        long bcnt = 0L;
	        int  mcnt = 0;
	        Message **mq = idToStation (i)->MQHead, *m;

	        display (i);
	        if (mq != NULL) {
	    	    for (m = mq [UTPattern->getId ()]; m != NULL; mcnt++,
			bcnt += m->Length, m = m->next);
	        }
	        display (mcnt); display (bcnt);
	    }
	}
}
#endif
