/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ----------------------------------------------------- */
/* Functions related to displaying windows on the screen */
/* ----------------------------------------------------- */

#include        <setjmp.h>
#include        "types.h"
#include        "var.h"
#include        "../SIMULATOR/display.h"

static  Window          *Selected = NULL;

static  jmp_buf         retusr;                 // Return from user input mode

extern  jmp_buf         restart;                // Return to main menu

void            selectFirst ();
void            moveToSelected ();

Rectangle::Rectangle (int xp, int yp, int xxp, int yyp) {

/* --------------------- */
/* Rectangle constructor */
/* --------------------- */

	x = xp; y = yp; xx = xxp; yy = yyp;

	next = FRectangles;
	FRectangles = this;
}

int     isFree (int x, int y, int xx, int yy, Window *ign) {

/* ---------------------------------------------------------- */
/* Checks   if   the   rectangle  bounded  by  the  specified */
/* coordinates does not itersect any existing rectangles      */
/* ---------------------------------------------------------- */

	Window  *w;

	for (w = Active; w != NULL; w = w->next) {
		if (w == ign) continue;
		if (w->x > xx) continue;
		if (w->y > yy) continue;
		if (w->x+w->xs <= x) continue;
		if (w->y+w->ys <= y) continue;
		return (NO);
	}
	return (YES);
}

static  int     isCont (int x, int y, int xx, int yy) {

/* ---------------------------------------------------------- */
/* Checks   if   the   rectangle  bounded  by  the  specified */
/* coordinates is  not  contained  within  an  already  found */
/* rectangle                                                  */
/* ---------------------------------------------------------- */

	Rectangle *r;

	for (r = FRectangles; r != NULL; r = r->next)
		if (r->x <= x && r->y <= y && r->xx >= xx && r->yy >= yy)
			return (YES);
	return (NO);
}

void    buildRectangles (int minw, int minh, Window *ign) {

/* ---------------------------------------------------------- */
/* Build  the set of free rectangles. minw and minh stand for */
/* the minimum width and height of a needed rectangle.        */
/* ---------------------------------------------------------- */

	short   *sx, *tx, *sy, *ty;
	int     nw, sxn, txn, syn, tyn, xx, yy, xxx, yyy, i, j, k, l;
	Window  *w;

	FRectangles = NULL;

	// Calculate the number of active windows
	for (nw = 0, w = Active; w != NULL; w = w->next)
		// Ignore the window to be ignored, if any
		if (w != ign) nw++;

	// Allocate arrays
	sx = new short [nw+1];
	tx = new short [nw+1];
	sy = new short [nw+1];
	ty = new short [nw+1];

	tx [0] = -1;            // Sentinels
	ty [0] = -1;
	sx [0] = NCols;
	sy [0] = NRows-1;

	for (w = Active, sxn = syn = txn = tyn = 1; w != NULL; w = w->next) {
		if (w == ign) continue; // A window to be ignored
		for (xx = w->x-1, i = 0; i < sxn && sx [i] < xx; i++);
		if (i == sxn) {
			sx [i] = xx;
			sxn++;
		} else if (xx != sx [i]) {
			for (j = sxn; j > i; j--) sx [j] = sx [j-1];
			sx [i] = xx;
			sxn++;
		}
		for (yy = w->y-1, i = 0; i < syn && sy [i] < yy; i++);
		if (i == syn) {
			sy [i] = yy;
			syn++;
		} else if (yy != sy [i]) {
			for (j = syn; j > i; j--) sy [j] = sy [j-1];
			sy [i] = yy;
			syn++;
		}
		for (xx = w->x + w->xs, i = 0; i < txn && tx [i] < xx; i++);
		if (i == txn) {
			tx [i] = xx;
			txn++;
		} else if (xx != tx [i]) {
			for (j = txn; j > i; j--) tx [j] = tx [j-1];
			tx [i] = xx;
			txn++;
		}
		for (yy = w->y + w->ys, i = 0; i < tyn && ty [i] < yy; i++);
		if (i == tyn) {
			ty [i] = yy;
			tyn++;
		} else if (yy != ty [i]) {
			for (j = tyn; j > i; j--) ty [j] = ty [j-1];
			ty [i] = yy;
			tyn++;
		}
	}

	// The loop below looks hopeless; however, in reality,
	// it executes reasonably fast

	minw--; minh--;
	if (nw == 0) {
		new Rectangle (-1, -1, NCols, NRows-1);
	} else {
		for (i = 0; i < txn; i++) {
			xx = tx [i];
			for (j = 0; j < sxn; j++) {
				if ((xxx = sx [j]) < xx + minw) continue;
				for (k = 0; k < tyn; k++) {
					yy = ty [k];
					for (l = 0; l < syn; l++) {
						if ((yyy = sy [l]) < yy + minh)
							continue;
						if (!isFree (xx, yy, xxx, yyy,
							ign))
								continue;
						if (isCont (xx, yy, xxx, yyy))
							continue;
						new Rectangle (xx,yy,xxx,yyy);
					}
					if (sy [syn-1] < yy + minh) break;
				}
			}
			if (sx [sxn-1] < xx + minw) break;
		}
	}

	delete (sx);
	delete (sy);
	delete (tx);
	delete (ty);
}

void    destroyRectangles () {

/* ------------------------- */
/* Clears the rectangle list */
/* ------------------------- */

	Rectangle  *r, *s;

	for (r = FRectangles; r != NULL; r = s) {
		s = r->next;
		delete (r);
	}

	FRectangles = NULL;
}

static  long    canFit (Rectangle* r, long width, long height, int &x, int &y,
			int space) {

/* ---------------------------------------------------------- */
/* Checks  if  the  rectangle is big enough to accomodate the */
/* window                                                     */
/* ---------------------------------------------------------- */

	long    w, h;

	if (space) {
		x = r->x + 1;
		y = r->y + 1;
		w = r->xx - x;
		h = r->yy - y;
	} else {
		if (r->x < 0) x = 0; else x = r->x;
		if (r->y < 0) y = 0; else y = r->y;
		w = r->xx - x + (r->xx != NCols);
		w = r->yy - y + (r->yy != NRows-1);
	}

	if (w >= width && h >= height)
		return ((w - width) * h + (h - height) * width);
	else
		return (NONE);
}

int     fitWindow (Template *t, int &x, int &y, int &xs, int &ys) {

/* ----------------------------- */
/* Allocates room for the window */
/* ----------------------------- */

	Rectangle *r, *sel;
	int       width, height, xci, yci, xp, yp, space;
	long      waste, ws;

	buildRectangles (((t->xclip == NULL) ? t->xclipd : t->xclip [0])+1,
			((t->yclip == NULL) ? t->yclipd : t->yclip [0])+1);
				// No window ignored

	if (FRectangles == NULL) return (NO);

	width = t->xclipd;
	height = t->yclipd;
	if (t->xclip != NULL) {
		for (xci = 0; xci < t->nxclip; xci++)
			if (t->xclip [xci] >= width) break;
	} else
		xci = 0;
	
	if (t->yclip != NULL) {
		for (yci = 0; yci < t->nyclip; yci++)
			if (t->yclip [yci] >= height) break;
	} else
		yci = 0;

	space = YES;

	// These are actually one unit too short
	width++;
	height++;

	while (1) {
		// Best fit leaving at least one space on each side not
		// adjacent to the screen boundary
		for (sel = NULL, waste = MAXlong, r = FRectangles; r != NULL;
			r = r->next) {

			if ((ws = canFit (r, width, height, xp, yp, space)) >= 0
				&& ws < waste) {

				sel = r;
				waste = ws;
				x = xp;
				y = yp;
			}
		}

		if (sel != NULL) {
			xs = width;
			ys = height;
			destroyRectangles ();
			return (YES);
		}

		if (xci == 0 && yci == 0) {
			// The last chance: try to fit the window without
			// leaving spaces on sides
			if (space) {
				space = NO;
				width = ((t->xclip == NULL) ? t->xclipd :
					t->xclip [0]) + 1;
				height = ((t->yclip == NULL) ? t->yclipd :
					t->yclip [0]) + 1;
				continue;
			}

			// No way
			destroyRectangles ();
			return (NO);
		}

		// Reduce size

		if (xci > 0) width = t->xclip [--xci] + 1;
		if (yci > 0) height = t->yclip [--yci] + 1;
	}
}

void    removeWindow (Window *w) {

/* ----------------------------------------- */
/* Erases the current window from the screen */
/* ----------------------------------------- */

	setCurrentWindow (w);
	for (int i = 0; i < w->ys; i++) blankOut (0, i, w->xs);

	// Check if this is the selected window
	if (w == Selected) {
		delete (w);
		Selected = NULL;
		selectFirst ();
	} else {
		delete (w);
	}
	moveToSelected ();
}

void    selectFirst () {

/* ---------------------------------------------------------- */
/* Selects  the  window  with  the smallest distance from the */
/* left to corner                                             */
/* ---------------------------------------------------------- */

	long    max, val;
	Window  *w;

	if (Active == NULL) return;

	for (max = MAXlong, w = Active; w != NULL; w = w->next)
		if ((val = w->x * w->x + w->y * w->y) < max) {
			max = val;
			Selected = w;
		}
}

char    *Window::makeHeader (int length, int dontfill) {

/* ---------------------------------------------------- */
/* Create the window header filling 'length' characters */
/* ---------------------------------------------------- */

	char    *cp, *cq;

	cp = CBuffer;
	if ((cq = objct->NName) == NULL) cq = objct->SName;
		
	while (length > 0 && *cq != '\0') {
		length--;
		*cp++ = *cq++;
	}

	if (length > 0) {
		*cp++ = ' ';
		length--;
	} else {
		*cp = '\0';
		return (CBuffer);
	}

	cq = form ("(%1d)", templ->mode);

	while (length > 0 && *cq != '\0') {
		length--;
		*cp++ = *cq++;
	}

	if (length <= 0) {
		*cp = '\0';
		return (CBuffer);
	}

	if (sttn != NONE) {
		cq = form ("/%1d", sttn);
		while (length > 0 && *cq != '\0') {
			length--;
			*cp++ = *cq++;
		}
	}

	if (!dontfill) {
		while (length > 0) {
			length--;
			*cp++ = ' ';
		}
	}
	*cp = '\0';

	return (CBuffer);
}

void    skipWindow () {

/* ---------------------------------------------------------- */
/* Ignores   the   rest   of   the  window  phrase  (somewhat */
/* heuristically, but no problems so far)                     */
/* ---------------------------------------------------------- */

	int     pk;

	while (1) {

		pk = peekOctet ();

		switch (pk) {

			case PH_EPH:
			case PH_REG:
			case PH_SCA:

				getOctet ();
				continue;

			case PH_INM:

				getInt ();
				continue;

			case PH_FNM:

				getFloat ();
				continue;

			case PH_BNM:

				getBIG ();
				continue;

			case PH_TXT:

				while (getOctet () != '\0');
				continue;

			case PH_SLI:

				getOctet ();
				getOctet ();
				getOctet ();
				continue;

			default:

				return;
		}
	}
}

char    *ltoa (long nn, char *s, int nc) {

/* ------------------------------------------- */
/* Encodes long integer with optional exponent */
/* ------------------------------------------- */

	int             i, j, jj, neg;
	long            al;
	char            *tm, *rs;

	tm = &(CBuffer [SANE-1]);
	rs = s;

	if (nn >= 0) {
		neg = 0;
		nn = -nn;
	} else
		neg = 1;


	for (j = 0; ; j++) {

		if (nn == 0) {
			if (j == 0) *--tm = '0'; else break;
			continue;
		}

		al = nn;
		*--tm = '0' - (char)(al - ((nn /= 10) * 10));
	}

	if ((jj=j+neg) <= nc) {
		while (nc-- >  jj) *s++ = ' ';
		if (neg) {
			*s++ = '-';
			nc--;
		}
		while (nc-- >= 0) *s++ = *tm++;
		*s = '\0';
		return (rs);
	}

	if ((i = jj - nc + 2) > 9)
		i = jj - nc + 3;

	if (i >= j) {
		while (nc-- > 0) *s++ = '*';
		*s = '\0';
		return (rs);
	}

	if (neg) *s++ = '-';
	while (j-- > i) *s++ = *tm++;
	*s++ = 'E';
	if (i > 9) {
		*s++ = (j = (i/10)) + '0';
		*s++ = (i - (j * 10)) + '0';
	} else {
		*s++ = i + '0';
	}
	*s ='\0';
	return (rs);
}

static int  storeItem (FilledField *r, char *cbuf) {

/* ---------------------------------------------- */
/* Stores a textual item in a filled field buffer */
/* ---------------------------------------------- */

	int     i, j;
	char    *cp, c;
	double  fn;
	BIG     bn;


	while (1) {
	 switch (peekOctet ()) {

	  case PH_INM :

		// Process integer number
		ltoa (getInt (), cbuf, r->length);
		if (r->just == LEFT) {
			for (cp = cbuf, i = 0; i < r->length && *cp == ' ';
				cp++, i++);
			if (i < r->length && cp != cbuf) {
				for ( ; i < r->length; i++) *cbuf++ = *cp++;
				while (cbuf < cp) *cbuf++ = ' ';
			}
		}
		return (0);

	  case PH_FNM :

		fn = getFloat ();
		if (r->length >= 7) {
			i = r->length - 6;
			if (fn >= 0.0) i++;
		} else
			i = 0;

		if (i > 3) i = 3;

		cp = form (form ("%%%1d.%1dg", r->length, i), fn);
		if (r->just == LEFT) {
			for (j = 0; j < r->length && *cp == ' '; cp++, j++);
			if (j >= r->length) {
				strncpy (cbuf, cp -= j, r->length);
			} else {
				for (i = j; i < r->length; i++) *cbuf++ = *cp++;
				while (j--) *cbuf++ = ' ';
			}
		} else {
			strncpy (cbuf, cp, r->length);
		}
		return (0);

	   case PH_BNM :

		bn = getBIG ();
		btoa (bn, cbuf, r->length);
		if (r->just == LEFT) {
			for (cp = cbuf, i = 0; i < r->length && *cp == ' ';
				cp++, i++);
			if (i < r->length && cp != cbuf) {
				for ( ; i < r->length; i++) *cbuf++ = *cp++;
				while (cbuf < cp) *cbuf++ = ' ';
			}
		}
		return (0);

	   case PH_TXT :

		getOctet ();
		for (cp = CBuffer, i = r->length; i > 0; i--) {
			if ((c = getOctet ()) == '\0') break;
			*cp++ = c;
		}
		*cp = '\0';
		if (c != '\0') while (getOctet () != '\0');

		if (r->just == RIGHT)
			while (i-- > 0) *cbuf++ = ' ';

		for (cp = CBuffer; (*cbuf++ = *cp++) != '\0'; );

		if (r->just == LEFT)
			while (i-- > 0) *cbuf++ = ' ';

		return (0);
			
	   case PH_EPH :

		getOctet ();
		continue;

	   default:

		// Assume that the display ends here
		return (1);
	 }
	}
}

static int  storeItem (RegionField *r, char *cbuf) {

/* ------------------------------------------------- */
/* Stores the contents of a region in a field buffer */
/* ------------------------------------------------- */

	float   x [SANE], y [SANE];
	double  xo, xs, yo, ys;

	int     np, i, width, height;
	long    bs, j;

	width = r->x1 - r->x + 1;
	height = r->y1 - r->y + 1;

	while ((i = peekOctet ()) == PH_EPH) getOctet ();
	if (i != PH_REG) {
		// Not a region item -- assume that the window ends here
		if (i == PH_INM || i == PH_FNM || i == PH_BNM || i == PH_TXT)
			abtDisplay (form ("storeItem: region expected", i));
		return (1);
	}

	getOctet ();    // Skip the starting byte

	if (peekOctet () == PH_SCA) {
		// Scaling present
		getOctet ();
		xo = getFloat ();
		xs = getFloat ();
		yo = getFloat ();
		ys = getFloat ();
	} else {
		xo = r->xo;
		xs = r->xs;
		yo = r->yo;
		ys = r->ys;
	}
	while (peekOctet () == PH_EPH) getOctet ();

	// Clear the buffer
	for (j = 0, bs = width * height; j < bs; j++) cbuf [j] = ' ';

	// Process segments
	while (peekOctet () == PH_SLI) {

		int     att, b, ix, iy;

		getOctet ();
		b = getOctet ();        // The attribute
		att = getOctet ();
		att |= b << 7;

		for (np = 0; np < SANE && peekOctet () == PH_FNM; np++) {
			x [np] = getFloat ();
			if (peekOctet () != PH_FNM) break;
			y [np] = getFloat ();
		}

		if (np == SANE)
			// Skip the rest
			while (peekOctet () == PH_FNM)
				getFloat ();

		// Process the segment
		if (np == 0) continue;

		// Check if should use default scaling
		if (xs == 0 || ys == 0) {
			// Default scaling needed -- compute min and max
			for (i = 0, yo = HUGE, ys = -HUGE; i < np; i++) {
				if (ys < y[i]) ys = y[i];
				if (yo > y[i]) yo = y[i];
			}
			if (ys == yo) {
				// Put it in the middle
				ys += ys/2.0;
				yo -= yo/2.0;
			}
			ys = ys - yo;

			// The same for the x coordinate
			for (i = 0, xo = HUGE, xs = -HUGE; i < np; i++) {
				if (xs < x[i]) xs = x[i];
				if (yo > x[i]) yo = x[i];
			}
			if (xs == xo) {
				// Put it in the middle
				xs += xs/2.0;
				xo -= xo/2.0;
			}
			xs = xs - xo;
		}

		for (i = 0; i < np; i++) {
			// Draw the points. Note: only points are supported
			// in this version (I don't think it makes much sense
			// to pretend to draw lines on the regular terminal
			// screen
			ix = (int)(((x[i] - xo)/xs) * width);
			iy = (int)(((y[i] - yo)/ys) * height);

			if (ix < 0) ix = 0; else if (ix >= width) ix=width-1;
			if (iy < 0) iy = 0; else if (iy >= height) iy=height-1;

			cbuf [ (height-iy-1) * width + ix ] = r->pc;
		}
		while (peekOctet () == PH_EPH) getOctet ();
	}
	return (0);
}

void    Window::displayNew () {

/* --------------------- */
/* Redisplays the window */
/* --------------------- */

	Field   *f;             FixedField      *fixf;
	char    *cp, *cq;       HeaderField     *hdrf;
				FilledField     *filf;
				RegionField     *regf;
				RepeatField     *repf;

	int     yinc,   // Current y increment for repetition
		crpc,   // Repetition counter

		erasing,        // Flag == erasing previous garbage

		row, cnt, yy;

	setCurrentWindow (this);

	erasing = NO;
	for (f = templ->fieldlist, cq = cp = contents, crpc = yinc = 0;
	    f != NULL;) {

		switch (f->type) {

		  case FT_HEADER:

			hdrf = (HeaderField*) f;
			if (hdrf->y + yinc >= ys) goto FORGET_IT;
			// Ignore fixed fields
			break;

		  case FT_FIXED:

			fixf = (FixedField*) f;
			if (fixf->y + yinc >= ys) goto FORGET_IT;
			break;

		  case FT_REGION:

			regf = (RegionField*) f;
			if (regf->y + yinc >= ys) goto FORGET_IT;
			if (erasing) {
				if (cp >= cq) goto FORGET_IT;
			} else {
				while (peekOctet () == PH_EPH) getOctet ();
				if (storeItem (regf, cp)) {
					// The window ends here, check for
					// erase needed
					cq = lff;
					lff = cp;
					if (cp < cq) {
						erasing = YES;
						while (cp < cq) *cp++ = ' ';
						cp = lff;
					} else {
						goto FORGET_IT;
					}
				}
			}
			// Draw the region
			cnt = regf->x1 - regf->x + 1;
			for (row = regf->y; row <= regf->y1; row++) {
				moveCursor (regf->x, row + yinc);
				outText (regf->dmode, cp, cnt);
				cp += cnt;
			}
			break;
				
		  case FT_FILLED:

			filf = (FilledField*) f;
			if ((yy = filf->y + yinc) >= ys) goto FORGET_IT;
			if (erasing) {
				if (cp >= cq) goto FORGET_IT;
			} else {
				while (peekOctet () == PH_EPH) getOctet ();
				if (storeItem (filf, cp)) {
					// The window ends here, check for
					// erase needed
					cq = lff;
					lff = cp;
					if (cp < cq) {
						erasing = YES;
						while (cp < cq) *cp++ = ' ';
						cp = lff;
					} else {
						goto FORGET_IT;
					}
				}
			}
			moveCursor (filf->x, yy);
			outText (filf->dmode, cp, filf->length);
			cp += filf->length;
			break;

		  case FT_REPEAT:

			if (yinc)
		       abtDisplay ("Illegal repetition field (internal error)");

			repf = (RepeatField*)f;
			f = repf->from;
			yinc = 1;
			crpc = 1;
			continue;
		  default:
			abtDisplay ("Illegal field type (internal error)");
		}

		if (yinc && f == repf->to) {
			if (repf->count != 0 && crpc >= repf->count) {
				// Switch to normal
				f = repf->next;
				if (f->type == FT_REPEAT) {
					repf = (RepeatField*)f;
					f = repf->from;
					crpc = 1;
					yinc++;
				} else {
					crpc = yinc = 0;
				}
				continue;
			}
			yinc++;
			crpc++;
			f = repf->from;
			continue;
		}

		f = f->next;
	}
	return;

FORGET_IT:

	if (!erasing) lff = cp;
	skipWindow ();  // Ignore the rest of the window phrase
}

void    Window::displayOld () {

/* --------------------- */
/* Redisplays the window */
/* --------------------- */

	Field   *f;             FixedField      *fixf;
	char    *cp;            HeaderField     *hdrf;
				FilledField     *filf;
				RegionField     *regf;
				RepeatField     *repf;

	int     yinc,   // Current y increment for repetition
		crpc,   // Repetition counter

		row, cnt, yy;

	setCurrentWindow (this);

	for (f = templ->fieldlist, cp = contents, crpc = yinc = 0; f != NULL;) {

		switch (f->type) {

		  case FT_HEADER:

			hdrf = (HeaderField*) f;
			if ((yy = hdrf->y + yinc) >= ys) return;
			moveCursor (hdrf->x, yy);
			outText (hdrf->dmode, makeHeader (hdrf->length));
			break;

		  case FT_FIXED:

			fixf = (FixedField*) f;
			if ((yy = fixf->y + yinc) >= ys) return;
			moveCursor (fixf->x, yy);
			outText (fixf->dmode, fixf->contents);
			break;

		  case FT_REGION:

			regf = (RegionField*) f;
			if (regf->y + yinc >= ys) return;
			cnt = regf->x1 - regf->x + 1;
			for (row = regf->y; row <= regf->y1; row++) {
				moveCursor (regf->x, row + yinc);
				outText (regf->dmode, cp, cnt);
				cp += cnt;
			}
			break;
				
		  case FT_FILLED:

			filf = (FilledField*) f;
			if ((yy = filf->y + yinc) >= ys) return;
			moveCursor (filf->x, yy);
			outText (filf->dmode, cp, filf->length);
			cp += filf->length;
			break;

		  case FT_REPEAT:

			if (yinc)
		       abtDisplay ("Illegal repetition field (internal error)");

			repf = (RepeatField*)f;
			f = repf->from;
			yinc = 1;
			crpc = 1;
			continue;
		  default:
			abtDisplay ("Illegal field type (internal error)");
		}

		if (yinc && f == repf->to) {
			if (repf->count != 0 && crpc >= repf->count) {
				// Switch to normal
				f = repf->next;
				if (f->type == FT_REPEAT) {
					repf = (RepeatField*)f;
					f = repf->from;
					crpc = 1;
					yinc++;
				} else {
					crpc = yinc = 0;
				}
				continue;
			}
			yinc++;
			crpc++;
			f = repf->from;
			continue;
		}

		f = f->next;
	}
}

long    Window::contentsLength () {

/* ----------------------------------------- */
/* Calculates the size of the contents array */
/* ----------------------------------------- */

	Field   *f;             FixedField      *fixf;
				HeaderField     *hdrf;
				FilledField     *filf;
				RegionField     *regf;
				RepeatField     *repf;

	int     yinc,   // Current y increment for repetition
		crpc,   // Repetition counter

		yy;

	long    bufl;

	for (f = templ->fieldlist, bufl = crpc = yinc = 0; f != NULL;) {

		switch (f->type) {

		  case FT_HEADER:

			hdrf = (HeaderField*) f;
			if ((yy = hdrf->y + yinc) >= ys) return (bufl);
			break;

		  case FT_FIXED:

			fixf = (FixedField*) f;
			if ((yy = fixf->y + yinc) >= ys) return (bufl);
			break;

		  case FT_REGION:

			regf = (RegionField*) f;
			if (regf->y + yinc >= ys) return (bufl);
			bufl += (regf->x1 - regf->x + 1) *
				(regf->y1 - regf->y + 1);
			break;
				
		  case FT_FILLED:

			filf = (FilledField*) f;
			if ((yy = filf->y + yinc) >= ys) return (bufl);
			bufl += filf->length;
			break;

		  case FT_REPEAT:

			if (yinc)
		       abtDisplay ("Illegal repetition field (internal error)");

			repf = (RepeatField*)f;
			f = repf->from;
			yinc = 1;
			crpc = 1;
			continue;
		  default:
			abtDisplay ("Illegal field type (internal error)");
		}

		if (yinc && f == repf->to) {
			if (repf->count != 0 && crpc >= repf->count) {
				// Switch to normal
				f = repf->next;
				if (f->type == FT_REPEAT) {
					repf = (RepeatField*)f;
					f = repf->from;
					crpc = 1;
					yinc++;
				} else {
					crpc = yinc = 0;
				}
				continue;
			}
			yinc++;
			crpc++;
			f = repf->from;
			continue;
		}

		f = f->next;
	}
	return (bufl);
}

int     Window::activeFields () {

/* ---------------------------------------------------- */
/* Calculates the number of active fields in the window */
/* ---------------------------------------------------- */

	Field   *f;             FixedField      *fixf;
				HeaderField     *hdrf;
				FilledField     *filf;
				RegionField     *regf;
				RepeatField     *repf;

	int     yinc,   // Current y increment for repetition
		crpc,   // Repetition counter

		yy;

	int     count;

	for (f = templ->fieldlist, count = crpc = yinc = 0; f != NULL;) {

		switch (f->type) {

		  case FT_HEADER:

			hdrf = (HeaderField*) f;
			if ((yy = hdrf->y + yinc) >= ys) return (count);
			break;

		  case FT_FIXED:

			fixf = (FixedField*) f;
			if ((yy = fixf->y + yinc) >= ys) return (count);
			break;

		  case FT_REGION:

			regf = (RegionField*) f;
			if (regf->y + yinc >= ys) return (count);
			count ++;
			break;
				
		  case FT_FILLED:

			filf = (FilledField*) f;
			if ((yy = filf->y + yinc) >= ys) return (count);
			count ++;
			break;

		  case FT_REPEAT:

			if (yinc)
		       abtDisplay ("Illegal repetition field (internal error)");

			repf = (RepeatField*)f;
			f = repf->from;
			yinc = 1;
			crpc = 1;
			continue;
		  default:
			abtDisplay ("Illegal field type (internal error)");
		}

		if (yinc && f == repf->to) {
			if (repf->count != 0 && crpc >= repf->count) {
				// Switch to normal
				f = repf->next;
				if (f->type == FT_REPEAT) {
					repf = (RepeatField*)f;
					f = repf->from;
					crpc = 1;
					yinc++;
				} else {
					crpc = yinc = 0;
				}
				continue;
			}
			yinc++;
			crpc++;
			f = repf->from;
			continue;
		}

		f = f->next;
	}
	return (count);
}

void    moveToSelected () {

/* --------------------------------------- */
/* Moves the cursor to the selected window */
/* --------------------------------------- */

	Window  *w;

	for (w = Active; w != NULL; w = w->next) if (w == Selected) break;
	if (w == NULL) {
		clearLLC ();
		absMove (0, 0);
		return;
	}
	outLLC (Selected->makeHeader (NCols, YES));
	absMove (Selected->x, Selected->y);
}

void    refreshScreen (int hold) {

/* ------------------- */
/* Rebuilds the screen */
/* ------------------- */

	Window  *w;
	int     i;

	clearScreen ();

	for (i = NO, w = Active; w != NULL; w = w->next) {
		w->displayOld ();
		if (w == Selected) i = YES;
	}

	if (!i) selectFirst ();
	if (Active == NULL)
		errMessage ("No windows to display");
	else if (hold)
		errMessage ("HOLDING");
	moveToSelected ();
}

static  void    Refresh () {

/* ------------------------------ */
/* Process the 'r' ('^L') command */
/* ------------------------------ */

	refreshScreen (NO);
	longjmp (retusr, 1);
}

static  void    Help () {

/* -------------------- */
/* Displays a help menu */
/* -------------------- */

	clearRLC ();                    // Cancel previous error message
	clearLLC ();
	clearScreen ();
	setCurrentWindow (0, 0, NCols, NRows);
	moveCursor (0, 0);
	outText (DM_REVERSE, "j/k ");
	outText (" moves the cursor to the next window down/up");
	moveCursor (0, 1);
	outText (DM_REVERSE, "l/h ");
	outText (" moves the cursor to the next window right/left");
	moveCursor (0, 2);
	outText (DM_REVERSE, "r,^L");
	outText (" refreshes the screen");
	moveCursor (0, 3);
	outText (DM_REVERSE, "w   ");
	outText (" exits to the object menu");
	moveCursor (0, 4);
	outText (DM_REVERSE, "x   ");
	outText (" exits to the startup menu");
	moveCursor (0, 5);
	outText (DM_REVERSE, "q   ");
	outText (" exits the program");
	moveCursor (0, 6);
	outText (DM_REVERSE, "m   ");
	outText (" moves the current window");
	moveCursor (0, 7);
	outText (DM_REVERSE, "c   ");
	outText (" clips (resizes) the current window");
	moveCursor (0, 8);
	outText (DM_REVERSE, "d   ");
	outText (" deletes the current window");
	moveCursor (0, 9);
	outText (DM_REVERSE, "s/u ");
	outText (" steps/unsteps the current window");
	moveCursor (0, 10);
	outText (DM_REVERSE, "U   ");
	outText (" unsteps all windows");
	moveCursor (0, 11);
	outText (DM_REVERSE, "Ret ");
	outText (" continues execution (advances to the next step)");
	moveCursor (0, 12);
	outText (DM_REVERSE, "Spc ");
	outText (" advances to the next step");
	moveCursor (0, 13);
	outText (DM_REVERSE, "i   ");
	outText (" changes the display interval");

	outLLC ("Hit any key to cancel the help menu");
	inChar ();
	clearLLC ();
	refreshScreen (YES);
}

void    Move () {

/* ------------------------ */
/* Moves the current window */
/* ------------------------ */

	Window    *w;
	Template  *t;
	char      c;
	int       cx, cy, cxx, cyy;
	Rectangle *r, *s;
	int       mw, mh, dist, d, i;

	clearRLC ();

	// Find the current window
	if (Active == NULL) {
		errMessage ("HOLDING");
		return;
	}
	for (w = Active; w != NULL; w = w->next) if (w == Selected) break;
	if (w == NULL)
		selectFirst ();

	outLLC ("h/j/k/l to the new location; Return to move, 'a' to abandon");

	absMove (Selected->x, Selected->y);
	cx = CurCol;
	cy = CurRow;

	// Build free rectangles

	t = Selected->templ;

	// Minimum width
	mw = Selected->xs;
	// Minimum height
	mh = Selected->ys;

	buildRectangles (mw, mh, Selected);

	while (1) {
		if ((c = inChar ()) == 'l') {
			if (cx + mw >= NCols) continue;
			cx++;
		} else if (c == 'h') {
			if (cx == 0) continue;
			cx--;
		} else if (c == 'j') {
			if (cy + mh >= NRows-1) continue;
			cy++;
		} else if (c == 'k') {
			if (cy == 0) continue;
			cy--;
		} else if (c == '\n') {
			break;
		} else if (c == 'a') {
			// Abandon
			goto M_EXIT;
		} else
			// Ignore
			continue;

		// Check if the new position is OK, i.e. if the current version
		// of the window will fit into a free square containing the
		// left upper corner

		cxx = cx + mw - 1;
		cyy = cy + mh - 1;

		// Check if the moved window would fit into one of free
		// rectangles

		if (!isCont (cx, cy, cxx, cyy)) {

			// No way: find the closest free rectangle big enough
			// according to the user-indicated direction and
			// move the cursor to its left upper corner

			for (r = FRectangles, dist = MAXint, s = NULL;
				r != NULL; r = r->next) {

				switch (c) {
				  case 'h':
					if (r->x >= cx) continue;
					break;
				  case 'l':
					if (r->x <= cx) continue;
					break;
				  case 'j':
					if (r->y <= cy) continue;
					break;
				  default:      // Should be 'k' (up)
					if (r->y >= cy) continue;
				}

				// If we are here, the rectangle is in the
				// right direction, check if it is big
				// enough

				if (r->xx - r->x < mw - 1 || r->yy - r->y <
					mh - 1) continue;       // Too small

				// Choose the closest one

				if ((d = (r->x - cx) * (r->x - cx) +
					(r->y - cy) * (r->y - cy)) < dist) {

					dist = d;
					s = r;
				}
			}

			if (s != NULL) {
				if ((cx = s->x) < 0) cx = 0;
				if ((cy = s->y) < 0) cy = 0;
			} else {
				// Stay where you are
				cx = CurCol;
				cy = CurRow;
			}
		}

		absMove (cx, cy);
	}

	// Perform the move operation

	if (Selected->x == cx && Selected->y == cy) goto M_EXIT;

	// Clear the present area

	setCurrentWindow (Selected);
	for (i = 0; i < Selected->ys; i++) blankOut (0, i, Selected->xs);

	Selected->x = cx;
	Selected->y = cy;

	// Display the window there
	Selected->displayOld ();

M_EXIT:

	destroyRectangles ();
	errMessage ("HOLDING");
	moveToSelected ();
}

void    Clip () {

/* --------------------------- */
/* Resizes the selected window */
/* --------------------------- */

	Window   *w;
	Template *t;
	short    xcl [SANE], ycl [SANE];
	int      nxcl, nycl, i, cxcl, cycl, tyc;
	char     c;

	clearRLC ();

	// Find the current window

	if (Active == NULL) {
		errMessage ("HOLDING");
		return;
	}

	for (w = Active; w != NULL; w = w->next) if (w == Selected) break;
	if (w == NULL)
		selectFirst ();

	// Copy clipping vectors

	if ((t = Selected->templ)->xclip != NULL) {
		for (i = 0; i < t->nxclip; i++)
			xcl [i] = t->xclip [i];
		nxcl = t->nxclip;
	} else {
		xcl [0] = t->xclipd;
		nxcl = 1;
	}

	if ((t = Selected->templ)->yclip != NULL) {
		for (i = 0; i < t->nyclip; i++)
			ycl [i] = t->yclip [i];
		nycl = t->nyclip;
	} else {
		ycl [0] = t->yclipd;
		nycl = 1;
	}

	// Start with the current clipping

	cxcl = Selected->xs - 1;        // Turn width and height to clipping
	cycl = Selected->ys - 1;        // column and row

outLLC ("h/j/k/l to the new LR corner; Return to resize, 'a' to abandon");

	absMove (Selected->x + cxcl, Selected->y + cycl);

	while (1) {

	  if ((c = inChar ()) == 'l') {

		// Move to the next clipping column

		for (i = 0; i < nxcl; i++) if (xcl [i] > cxcl) break;
		if (i >= nxcl) continue;
		if (Selected->x + xcl [i] >= NCols) continue;
		
		if (!isFree (Selected->x, Selected->y, Selected->x + xcl[i],
			Selected->y + cycl, Selected)) continue;
			// Note: the resized window is ignored

		cxcl = xcl [i];

	  } else if (c == 'h') {

		// Move to the previous clipping column
		for (i = 0; i < nxcl; i++) if (xcl [i] >= cxcl) break;
		if (--i < 0) continue;
		cxcl = xcl [i];

	  } else if (c == 'j') {

		// Move to the next clipping row

		for (i = 0; i < nycl; i++) if (ycl [i] > cycl) break;

		if (i < nycl)
			tyc = ycl [i];
		else if (t->unbounded)
			tyc = cycl + 1;
		else
			continue;

		if (Selected->y + tyc >= NRows - 1) continue;
		
		if (!isFree (Selected->x, Selected->y, Selected->x + cxcl,
			Selected->y + tyc, Selected)) continue;

		cycl = tyc;

	  } else if (c == 'k') {

		// Move to the previous clipping column
		if (t->unbounded && cycl > ycl [nycl - 1]) {
			cycl--;
		} else {
			for (i = 0; i < nycl; i++) if (ycl [i] >= cycl) break;
			if (--i < 0) continue;
			cycl = ycl [i];
		}
	  } else if (c == '\n') {
		break;
	  } else if (c == 'a') {
		goto C_EXIT;
	  } else
		continue;

	  absMove (Selected->x + cxcl, Selected->y + cycl);
	}

	if (Selected->xs != cxcl+1 || Selected->ys != cycl+1) {

		// The size has changed
		setCurrentWindow (Selected);
		for (int i = 0; i < Selected->ys; i++)
				blankOut (0, i, Selected->xs);

		Selected->xs = cxcl+1;
		Selected->ys = cycl+1;

		setTimer ();
		sendOctet (PH_WND);
		sendText (Selected->objct->SName);
		sendInt (t->mode);
		// Should be station relative ?
		if (Selected->sttn >= 0) sendInt (Selected->sttn);
		sendOctet (PH_TRN);
		sendInt (Selected->activeFields ());
		sendOctet (PH_EPH);
		clearTimer ();

		// Reallocate contents
		delete (Selected->contents);
		// Re-calculate the size of the contents array
		if ((tyc = (int)(Selected->contentsLength ())) > 0) {
			Selected->contents = new char [tyc+1];
			for (i = 0; i < tyc; i++) Selected->contents [i] = ' ';
		} else
			Selected->contents = NULL;
		Selected->lff = Selected->contents;

		Selected->displayOld ();
	}

C_EXIT:
	errMessage ("HOLDING");
	moveToSelected ();
}

void    Step () {

/* --------------------------------------- */
/* Puts the selected window into step mode */
/* --------------------------------------- */

	Window    *w;
	char      *st;
	int       ex;
	BIG       at;

	clearRLC ();

	// Find the current window

	if (Active == NULL) {
		errMessage ("HOLDING");
		return;
	}

	for (w = Active; w != NULL; w = w->next) if (w == Selected) break;
	if (w == NULL) selectFirst ();
	
	st = inNumber ("At: IMMEDIATELY", 4, ex);
	clearRLC ();

	if (ex == NONE) goto S_EXIT;

	setTimer ();
	sendOctet (PH_WND);
	sendText (Selected->objct->SName);
	sendInt (Selected->templ->mode);
	// Should be station relative ?
	if (Selected->sttn >= 0) sendInt (Selected->sttn);

	sendOctet (PH_STP);

	if (ex == YES)          // Delayed
		sendBIG (ctob (st));

	sendOctet (PH_EPH);
	clearTimer ();

S_EXIT:

	errMessage ("HOLDING");
	moveToSelected ();
}

void    UnStep () {

/* --------------------------- */
/* Unsteps the selected window */
/* --------------------------- */

	Window    *w;

	clearRLC ();

	// Find the current window

	if (Active == NULL) {
		errMessage ("HOLDING");
		return;
	}

	for (w = Active; w != NULL; w = w->next) if (w == Selected) break;
	if (w == NULL) selectFirst ();
	
	setTimer ();
	sendOctet (PH_WND);
	sendText (Selected->objct->SName);
	sendInt (Selected->templ->mode);
	// Should be station relative ?
	if (Selected->sttn >= 0) sendInt (Selected->sttn);

	sendOctet (PH_STP);
	sendOctet (PH_ESM);
	sendOctet (PH_EPH);
	clearTimer ();

	errMessage ("HOLDING");
	moveToSelected ();
}

static  void    UnStepAll () {

/* ------------------ */
/* Unsteps everything */
/* ------------------ */

	clearRLC ();

	setTimer ();
	sendOctet (PH_ESM);
	sendOctet (PH_EPH);
	clearTimer ();
	errMessage ("HOLDING");
	moveToSelected ();
}
	
void    Updd () {

/* --------------------------- */
/* Moves to the next window Up */
/* --------------------------- */

	Window  *w, *v;
	int     dist;

	clearRLC ();
	// Find the current window

	for (w = Active; w != NULL; w = w->next) if (w == Selected) break;
	if (w == NULL) {
		selectFirst ();
		errMessage ("HOLDING");
		moveToSelected ();
		return;
	}

	// Find the closest window up
	for (dist = MAXint, w = Active, v = NULL; w != NULL; w = w->next) {

		int     d;

		if (w == Selected) continue;
		if (w->y >= Selected->y) continue;
		if ((d = (w->x - Selected->x) * (w->x - Selected->x) +
			(w->y - Selected->y) * (w->y - Selected->y)) < dist) {

			dist = d;
			v = w;
		}
	}

	if (v != NULL) {
		Selected = v;
		errMessage ("HOLDING");
		moveToSelected ();
	}
}

void    Downdd () {

/* ----------------------------- */
/* Moves to the next window down */
/* ----------------------------- */

	Window  *w, *v;
	int     dist;

	clearRLC ();
	// Find the current window

	for (w = Active; w != NULL; w = w->next) if (w == Selected) break;
	if (w == NULL) {
		selectFirst ();
		errMessage ("HOLDING");
		moveToSelected ();
		return;
	}

	// Find the closest window down
	for (dist = MAXint, w = Active, v = NULL; w != NULL; w = w->next) {

		int     d;

		if (w == Selected) continue;
		if (w->y <= Selected->y) continue;
		if ((d = (w->x - Selected->x) * (w->x - Selected->x) +
			(w->y - Selected->y) * (w->y - Selected->y)) < dist) {

			dist = d;
			v = w;
		}
	}

	if (v != NULL) {
		Selected = v;
		errMessage ("HOLDING");
		moveToSelected ();
	}
}

void    Leftdd () {

/* ----------------------------- */
/* Moves to the next window left */
/* ----------------------------- */

	Window  *w, *v;
	int     dist;

	// Find the current window

	clearRLC ();
	for (w = Active; w != NULL; w = w->next) if (w == Selected) break;
	if (w == NULL) {
		selectFirst ();
		errMessage ("HOLDING");
		moveToSelected ();
		return;
	}

	// Find the closest window down
	for (dist = MAXint, w = Active, v = NULL; w != NULL; w = w->next) {

		int     d;

		if (w == Selected) continue;
		if (w->x >= Selected->x) continue;
		if ((d = (w->x - Selected->x) * (w->x - Selected->x) +
			(w->y - Selected->y) * (w->y - Selected->y)) < dist) {

			dist = d;
			v = w;
		}
	}

	if (v != NULL) {
		Selected = v;
		errMessage ("HOLDING");
		moveToSelected ();
	}
}

void    Rightdd () {

/* ------------------------------ */
/* Moves to the next window right */
/* ------------------------------ */

	Window  *w, *v;
	int     dist;

	clearRLC ();
	// Find the current window

	for (w = Active; w != NULL; w = w->next) if (w == Selected) break;
	if (w == NULL) {
		selectFirst ();
		errMessage ("HOLDING");
		moveToSelected ();
		return;
	}

	// Find the closest window down
	for (dist = MAXint, w = Active, v = NULL; w != NULL; w = w->next) {

		int     d;

		if (w == Selected) continue;
		if (w->x <= Selected->x) continue;
		if ((d = (w->x - Selected->x) * (w->x - Selected->x) +
			(w->y - Selected->y) * (w->y - Selected->y)) < dist) {

			dist = d;
			v = w;
		}
	}

	if (v != NULL) {
		Selected = v;
		errMessage ("HOLDING");
		moveToSelected ();
	}
}

void    Remove () {

/* -------------------------- */
/* Removes the current window */
/* -------------------------- */

	Window  *w;

	clearRLC ();
	// Find the current window

	for (w = Active; w != NULL; w = w->next) if (w == Selected) break;
	if (w == NULL) {
		selectFirst ();
		errMessage ("HOLDING");
		moveToSelected ();
		return;
	}
	removeWindow (w);
	setTimer ();
	sendOctet (PH_EWN);
	sendText (w->objct->SName);
	sendInt (w->templ->mode);
	// Should be station relative ?
	if (w->sttn >= 0) sendInt (w->sttn);
	sendOctet (PH_EPH);
	clearTimer ();
	errMessage ("HOLDING");
}

static  void    Interval () {

/* ------------------------- */
/* Redefine display interval */
/* ------------------------- */

	char    *st;
	int     ex;

	clearRLC ();
	while (1) {
		st = "Interval: 5000";
		st = inNumber (st, 10, ex);
		clearRLC ();
		if (ex == NONE) {
			// Aborted
			moveToSelected ();
			return;
		} else if (ex == YES) {
			ex = atoi (st);
			// Check if legal
			if (ex <= 0) {
				errMessage ("Illegal display interval");
				continue;
			}
		} else {
			ex = 5000;
		}
		break;
	}
	setTimer ();
	sendOctet (PH_DIN);
	sendInt (ex);
	sendOctet (PH_EPH);
	clearTimer ();
	moveToSelected ();
	errMessage ("HOLDING");
}

static  void    Terminate () {

/* --------------------- */
/* Terminates the smurph */
/* --------------------- */

	errMessage ("One more 't' to terminate, any other key to ignore");
	if (inChar () == 't') {
		setTimer ();
		sendOctet (PH_EXI);
		sendOctet (PH_EPH);
		clearTimer ();
		TermFlag = YES;
	}
	errMessage ("HOLDING");
	moveToSelected ();
}

static  void    Quit () {

/* ---------------------- */
/* Terminates the program */
/* ---------------------- */

	clearScreen ();
	closeDisplay ();
	exit (0);
}

static  void    Other (char c) {

/* -------------------------- */
/* Process special characters */
/* -------------------------- */

	switch (c) {

		case 't':

			// Terminate smurph
			Terminate ();
			return;

		case 'w':

			// Menu
			longjmp (retusr, 2);

		case 'm':

			Move ();
			return;

		case 'c':

			Clip ();
			return;

		case 's':

			Step ();
			return;

		case 'd':

			Remove ();
			return;

		case 'i':

			Interval ();
			return;

		case 'u':

			UnStep ();
			return;

		case 'U':

			UnStepAll ();
			return;

		case '\n':

			clearRLC ();
			longjmp (retusr, 1);    // Proceed

		case ' ':

			if (Stepped == YES) longjmp (retusr, 1);
			return;

		default:

			errMessage ("HOLDING");
	}
}

static  void    Exit () {

/* ----------- */
/* Process 'x' */
/* ----------- */

	longjmp (restart, 1);
}

static  int     userInput () {

/* ---------------------------------------------- */
/* Processes the user command in the display mode */
/* ---------------------------------------------- */

	int     k;

	// Set up action functions

	clearActions ();

	ScRefresh = Refresh;
	ScUp = Updd;
	ScDown = Downdd;
	ScLeft = Leftdd;
	ScRight = Rightdd;
	ScQuit = Quit;
	ScExit = Exit;
	ScHelp = Help;
	ScOther = Other;

	if (k = setjmp (retusr)) return (k-1);

	processUserInput ();

	return (0);
}

int     displayWindows () {

/* ------------------------------------------ */
/* Prepares the screen for displaying windows */
/* ------------------------------------------ */

	NewWindow  *u, *v;

	WCNAdmitted = YES;
	TermFlag = NO;

	// At this moment smurph is awaiting a response

	if (Updates != NULL) {
		for (u = Updates; u != NULL; u = v) {
			v = u->next;
			u->wind->requestDisplay ();
			delete (u);
		}
		refreshScreen (YES);
		Updates = NULL;
		if (userInput ()) return (0);   // Menu requested
		setTimer ();
		sendOctet (PH_ESP);
		clearTimer ();
		clearRLC ();
	} else {
		setTimer ();
		sendOctet (PH_ESP);
		clearTimer ();
		refreshScreen (NO);
	}

	// Wait for the user's and smurph's input at the same time

	while (1) {
	    if (readyParty () == IO_USER) {
		// Terminal input
		setTimer ();
		sendOctet (PH_ATT);             // Force smurph to respond
		readSmurphSpurt ();
		clearTimer ();
		if (Stepped == NONE) {
		   if (!TermFlag) {
			outRLC ("Execution terminated: 'return' to exit");
			while (inChar () != '\n');
		   }
		   sleep (2);                   // Give it time to disappear
		   return (1);
		}
		if (userInput ()) return (0);   // Menu requested
		clearRLC ();
		sendOctet (PH_ESP);             // Let it continue
	    } else {
		// Smurph's spurt
		setTimer ();
		readSmurphSpurt ();
		if (Stepped >= YES) {
			// Stepped window updated or locked on PH_LCK
			clearTimer ();
			if (Stepped == YES) outRLC ("STEP");
			if (userInput ()) return (0);
			setTimer ();
		} else if (Stepped == NONE) {
			// Execution terminated
			clearTimer ();
			if (!TermFlag) {
			    outRLC ("Execution terminated: 'return' to exit");
			    while (inChar () != '\n');
			}
			sleep (2);
			return (1);
		}
		sendOctet (PH_ESP);
		clearTimer ();
	    }
	}
}
