/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ------------------------------------ */
/* Functions processing the object menu */
/* ------------------------------------ */

#include        <setjmp.h>
#include        "types.h"
#include        "var.h"

static  jmp_buf done;                   // The return context
static  TItem   *Selected = NULL,       // The selected object
		*Top,                   // The top object on the screen
		*Bottom,                // The bottom object on the screen
		*Owner;                 // The current parent

static  int     SHeight,                // Number of usable rows on the screen
		ItemSize,
		NItemsInLine;

static  void    Refresh ();
static  void    computeLayoutF (TItem*);
static  void    computeLayoutB (TItem*);
static  void    computeItemSize ();
static  void    dispScreen ();
static  void    cursorToSelected ();

Window::Window (TItem *ob, Template *tm, int xp, int yp, int xps, int yps, 
			int st) {

/* ---------------------- */
/* The window constructor */
/* ---------------------- */

	int             rpt, contl;

	templ = tm;
	x = xp;
	y = yp;
	xs = xps;
	ys = yps;
	objct = ob;
	sttn = st;

	// Add it to the list
	if ((next = Active) != NULL) next->prev = this;
	Active = this;
	prev = NULL;

	// And to the list of updates
	new NewWindow (this);

	// Calculate the size of the contents array
	if ((contl = (int) contentsLength ()) > 0) {
		contents = new char [contl+1];
		for (rpt = 0; rpt < contl; rpt++) contents [rpt] = ' ';
	} else
		contents = NULL;
	lff = contents;
}

Window::~Window () {

/* --------------------- */
/* The window destructor */
/* --------------------- */

	if (prev == NULL)
		Active = next;
	else
		prev->next = next;

	if (next != NULL)
		next->prev = prev;

	if (contents != NULL) delete (contents);
}

NewWindow::NewWindow (Window *w) {

/* ------------------------------ */
/* Constructor for an update item */
/* ------------------------------ */

	wind = w;
	next = Updates;
	Updates = this;
}

static  void    Quit () {

/* ------------------- */
/* Terminate execution */
/* ------------------- */

	clearScreen ();
	closeDisplay ();
	exit (0);
}

static  void    Exit () {

/* --------------------- */
/* Exit the current mode */
/* --------------------- */

	longjmp (done, 1);
}

static  void    Help () {

/* -------------------- */
/* Displays a help menu */
/* -------------------- */

	clearRLC ();                    // Cancel previous error message
	clearLLC ();
	clearScreen ();
	refreshCorners ();

	moveCursor (0, 0);
	outText (DM_REVERSE, "j/k ");
outText (" moves the cursor to the first item in the next/previous line");
	moveCursor (0, 1);
	outText (DM_REVERSE, "l/h ");
	outText (" moves the cursor to the next/previous item");
	moveCursor (0, 2);
	outText (DM_REVERSE, "n/p ");
	outText (" moves to the next/previous screen");
	moveCursor (0, 3);
	outText (DM_REVERSE, "r,^L");
	outText (" refresh the screen");
	moveCursor (0, 4);
	outText (DM_REVERSE, "q   ");
	outText (" exits the program");
	moveCursor (0, 5);
	outText (DM_REVERSE, "x   ");
	outText (" exits the menu");
	moveCursor (0, 6);
	outText (DM_REVERSE, "a   ");
	outText (" adds the window to the active list");
	moveCursor (0, 7);
	outText (DM_REVERSE, "c   ");
	outText (" moves to the selected child");
	moveCursor (0, 8);
	outText (DM_REVERSE, "f   ");
	outText (" moves to the parent of the current 'owner'");

	outLLC ("Hit any key to proceed");
	inChar ();
	clearLLC ();
	Refresh ();
}

static  void    Down () {

/* -------------------------------------------------- */
/* Move the cursor to the first item of the next line */
/* -------------------------------------------------- */

	TItem   *it;
	int     k, x, y;

	clearRLC ();
	if (Selected == Owner) {
		Selected = Top;
		cursorToSelected ();
		return;
	}

	it = Top;
	k = 0;                  // Row index
	y = 0;                  // Column index

	// Locate the current selection

	while (1) {
		x = k * (ItemSize + 2);
		if (it == Selected || it == Bottom) break;
		it = it->lnext;
		if (++k >= NItemsInLine) {
			k = 0;
			y++;
		}
	}

	if (it != Selected) {
		// This shouldn't happen
		it = Selected = Top;
		x = 0; y = 0; k = 0;
	}

	while (it != Bottom) {
		it = it->lnext;
		if (++k >= NItemsInLine) {
			Selected = it;
			cursorToSelected ();
			return;
		}
	}
	errMessage ("At the bottom");
	cursorToSelected ();
}

static  void    Up () {

/* ------------------------------------------------------ */
/* Move the cursor to the first item of the previous line */
/* ------------------------------------------------------ */

	TItem   *it;
	int     k, x, y;

	clearRLC ();
	if (Selected == Owner) {
		errMessage ("At the top");
		cursorToSelected ();
		return;
	}

	it = Top;
	k = 0;                  // Row index
	y = 0;                  // Column index

	// Locate the current selection

	while (1) {
		x = k * (ItemSize + 2);
		if (it == Selected || it == Bottom) break;
		it = it->lnext;
		if (++k >= NItemsInLine) {
			k = 0;
			y++;
		}
	}

	if (it != Selected) {
		// This shouldn't happen
		Selected = Owner;
		cursorToSelected ();
		return;
	}

	while (it != Top) {
		it = it->lprev;
		if (k-- == 0) {
			for (k = NItemsInLine-1; k; k--) it = it->lprev;
			Selected = it;
			cursorToSelected ();
			return;
		}
	}

	Selected = Owner;
	cursorToSelected ();
	return;
}

static  void    Right () {

/* -------------------------------- */
/* Move the cursor to the next item */
/* -------------------------------- */

	clearRLC ();
	if (Selected == Owner) {
		Selected = Top;
		cursorToSelected ();
		return;
	}

	if (Selected != Bottom)
		Selected = Selected->lnext;
	else
		errMessage ("No next object");
	cursorToSelected ();
}

static  void    Left () {

/* ------------------------------------ */
/* Move the cursor to the previous item */
/* ------------------------------------ */

	clearRLC ();
	if (Selected != Owner) {
		if (Selected == Top)
			Selected = Owner;
		else
			Selected = Selected->lprev;
	} else
		errMessage ("No previous object");
	cursorToSelected ();
}

static  void    Next () {

/* ----------------------- */
/* Move to the next screen */
/* ----------------------- */

	clearRLC ();
	if (Bottom->lnext != NULL) {

		computeLayoutF (Bottom->lnext);
		Selected = Top;
		Refresh ();
	} else {
		errMessage ("No more objects");
		cursorToSelected ();
	}
}

static  void    Prev () {

/* --------------------------- */
/* Move to the previous screen */
/* --------------------------- */

	clearRLC ();
	if (Top->lprev != (TItem*)(&(Owner->ChList))) {

		computeLayoutB (Top->lprev);
		Selected = Top;
		Refresh ();
	} else {
		errMessage ("No previous screen");
		cursorToSelected ();
	}
}

static  void    Parent () {

/* ------------------ */
/* Move to the parent */
/* ------------------ */

	TItem   *t;

	clearRLC ();
	if (Owner != Tree && (t = findParent (Owner)) != NULL) {
		Owner = t;
		computeItemSize ();
		computeLayoutF (Owner->ChList);
		Selected = Owner;
		clearScreen ();
		outLUC (Owner->dName ());
		outRUC ("Hit '?' for help");
		outLLC (Owner->dName ());
		dispScreen ();
	} else {
		errMessage ("This is the root");
		cursorToSelected ();
	}
}

static  void    Child () {

/* ----------------- */
/* Move to the child */
/* ----------------- */

	TItem   *t;

	clearRLC ();

	// Find the current selection

	for (t = Top; t != NULL; t = t->lnext)
		if (t == Selected || t == Bottom) break;

	if (t != Selected) {
		Selected = Owner;
		cursorToSelected ();
		return;
	}
		
	if (Selected->ChList == NULL) {
		errMessage ("No children");
		cursorToSelected ();
	} else {
		Owner = Selected;
		computeItemSize ();
		computeLayoutF (Owner->ChList);
		clearScreen ();
		outLUC (Owner->dName ());
		outRUC ("Hit '?' for help");
		outLLC (Owner->dName ());
		dispScreen ();
	}
}

Template *findTemp (TItem *obj, int &mode, int station) {

/* -------------------------------------------------------------- */
/* Finds the template for the given object and the indicated mode */
/* -------------------------------------------------------------- */

	Template *tm;
	char     *tn;
	int      fmode;

	tn = obj->tName (obj->SName);
	fmode = NONE;
	for (int k = 0; k < 2; k++) {
	    for (tm = TempList; tm != NULL; tm = tm->next) {
		if (strcmp (tm->typename, tn) == 0) {
			if (mode == -2) {
				// Request to determine whether there is more
				// than one mode
				if (fmode != NONE && tm->mode != fmode)
					// Affirmative
					return (tm);
				fmode = tm->mode;
			} else if (mode == NONE) {
				// Looking for the first mode
				mode = tm->mode;
				return (tm);
			} else {
				// A specific mode sought
				if (mode == tm->mode) {
					if (station == -2 ||
					    tm->status == TS_FLEXIBLE ||
			(station == NONE && tm->status == TS_ABSOLUTE) ||
			(station >= 0 && tm->status == TS_RELATIVE))
					return (tm);
				}
			}
		}
	    }
	    // No luck -- try the standard type name
	    if ((tn = obj->tName (obj->BName)) == NULL) break;
	}
	return (NULL);

}

int tempStatus (TItem *obj, int mode) {

/* ----------------------------------------------- */
/* Returns the station-relative status of the mode */
/* ----------------------------------------------- */

	Template *tm;
	char     *tn;
	int      stat;

	tn = obj->tName (obj->SName);
	stat = 0;
	for (int k = 0; k < 2; k++) {
	    for (tm = TempList; tm != NULL; tm = tm->next) {
		if (strcmp (tm->typename, tn) == 0 && tm->mode == mode) {
			switch (tm->status) {
				case TS_ABSOLUTE:
					if (stat == TS_ABSOLUTE) continue;
					if (stat == 0) {
						stat = TS_ABSOLUTE;
						continue;
					}
					return (TS_FLEXIBLE);
				case TS_FLEXIBLE:
					return (TS_FLEXIBLE);
				default:
					if (stat == 0) {
						stat = TS_RELATIVE;
						continue;
					}
					return (TS_FLEXIBLE);
			}
		}
	    }
	    if ((tn = obj->tName (obj->BName)) == NULL) break;
	}
	if (stat == 0) stat = TS_ABSOLUTE;
	return (stat);
}

void    Add () {

/* ----------------------------------------------------- */
/* Adds the current window to the list of active windows */
/* ----------------------------------------------------- */

	Template   *tm;
	char       *st;
	int        mode, md, ex, stat, x, y, xs, ys;
	Window     *w;

	clearRLC ();

	mode = NONE;
	if ((tm = findTemp (Selected, mode)) == NULL) {
		errMessage ("No template for this object");
		cursorToSelected ();
		return;
	}

	// Note: we are stopped at the first template matching the object's
	//       type name

	md = -2;
	// Check if there are templates for more than one mode
	if (findTemp (Selected, md) != NULL) {
		// More than one mode, prompt the user for the mode
		clearLLC ();            // Will restore later
		clearRUC ();            // '?' won't work
ARTR:
		// Include the default (first mode found)
		st = form ("Mode: %d", mode);
		st = inNumber (st, 6, ex);
		clearRLC ();
		if (ex == NONE) {
			// Aborted
AABT:                   outRUC ("Hit '?' for help");
			cursorToSelected ();
			return;
		} else if (ex == YES) {
			mode = atoi (st);
			// Check if legal
			if (findTemp (Selected, mode) == NULL) {
				errMessage ("Illegal mode");
				mode = tm->mode;
				goto ARTR;      // Retry
			}
		}                       // Default otherwise
	}

	// Now check if the window should be station relative
	stat = tempStatus (Selected, mode);
	if (stat != TS_ABSOLUTE) {
		clearLLC ();
SRTR:
		if (stat == TS_FLEXIBLE) {
			st = form ("Station: NONE");
			st = inNumber (st, 9, ex);
		} else if (stat == TS_RELATIVE) {
			st = inNumber ("Station: ", 9, ex);
		}
		clearRLC ();
		if (ex == NONE) goto AABT;
		if (ex == NO) {
			if (stat == TS_RELATIVE) {
				errMessage ("Station must be specified");
				goto SRTR;      // Retry
			}
			stat = TS_ABSOLUTE;
		} else {
			int tstat = atoi (st);
			if (tstat >= NStations) {
				errMessage ("Illegal station number");
				goto SRTR;
			}
			stat = tstat;
		}
	}

	// OK, find the right template

	if (stat < 0) stat = NONE;
	if ((tm = findTemp (Selected, mode, stat)) == NULL)
		abtDisplay ("Can't find template (internal error)");

	// Check if the window is not already displayed

	for (w = Active; w != NULL; w = w->next) {
		if (w->templ == tm && w->objct == Selected && w->sttn == stat) {
			// Already defined
			errMessage ("Window already active");
			goto AABT;
		}
	}

	// Find the room for the window on the screen

	if (fitWindow (tm, x, y, xs, ys) == NO) {
		errMessage ("Not enough room on the screen");
		goto AABT;
	}

	new Window (Selected, tm, x, y, xs, ys, stat);
	outRUC ("Hit '?' for help");
	cursorToSelected ();
	return;
}

static  void    Other   (char c) {

/* -------------------- */
/* Processes other keys */
/* -------------------- */

	switch (c) {

		case 'f' : Parent (); break;
		case 'c' : Child (); break;
		case 'a' : Add (); break;

		default: errMessage ("Undefined key");
			 cursorToSelected ();
			 return;

	}
}

static  void    cursorToSelected () {

/* ------------------------------------- */
/* Moves the cursor to the selected item */
/* ------------------------------------- */
	
	TItem   *it;
	int     k, x, y;

	if (Selected == Owner) {
		outLLC (Owner->dName ());
		absMove (0, 0);
	} else {
		it = Top;
		k = 0;                  // Row index
		y = 0;                  // Column index

		while (1) {
			x = k * (ItemSize + 2);
			if (it == Selected || it == Bottom) break;
			it = it->lnext;
			if (++k >= NItemsInLine) {
				k = 0;
				y++;
			}
		}

		if (it != Selected) {
			// This shouldn't happen
			Selected = Top;
			x = 0; y = 0;
		}

		outLLC (Selected->dName ());
		moveCursor (x, y);
	}
}

static  void    dispScreen () {

/* -------------------------------------- */
/* Displays the object menu on the screen */
/* -------------------------------------- */

	TItem   *it;
	int     k, n;

	it = Top;
	k = 0;                  // Row index
	n = 0;                  // Column index

	while (1) {
		moveCursor (k * (ItemSize + 2), n);
		outText (it->dName ());
		if (it == Bottom) break;
		it = it->lnext;
		if (++k >= NItemsInLine) {
			k = 0;
			n++;
		}
	}

	cursorToSelected ();     // Move the cursor to the selected item
}

static  void    Refresh () {

/* -------------------- */
/* Refreshes the screen */
/* -------------------- */

	clearRLC ();            // Error message
	clearLLC ();
	clearScreen ();
	refreshCorners ();
	dispScreen ();
}

static  void    computeItemSize () {

/* ---------------------------------------------------------- */
/* Determines  how many items will fit into a single line and */
/* how much space each item takes                             */
/* ---------------------------------------------------------- */

	TItem   *t;
	int     n;

	// Compute the length of the longest item on the list

	for (ItemSize = 0, t = Owner->ChList; t != NULL; t = t->lnext)
		if ((n = strlen (t->dName ())) > ItemSize) ItemSize = n;

	// Now, see how many of them can be fit into one row, so that
	// at least two blanks separate consecutive items

	NItemsInLine = 1;
	if (NCols <= ItemSize)
		ItemSize = NCols;
	else
		NItemsInLine += (NCols - ItemSize) / (ItemSize + 2);
}

static  void    computeLayoutF (TItem *il) {

/* ----------------------------------------- */
/* Determines how to put items on the screen */
/* ----------------------------------------- */

	int     n;

	n = SHeight * NItemsInLine;

	for (Bottom = Top = il; --n > 0 && Bottom->lnext != NULL; Bottom =
		Bottom->lnext);
}

static  void    computeLayoutB (TItem *il) {

/* -------------------------------- */
/* The same as above, but backwards */
/* -------------------------------- */

	int     n;

	n = SHeight * NItemsInLine;

	for (Bottom = Top = il; --n > 0 &&
		Top->lprev != (TItem*)(&(Owner->ChList)); Top = Top->lprev);
}

void    objectMenu () {

/* --------------------------------------------------- */
/* The driver function controlling the menu processing */
/* --------------------------------------------------- */

	clearScreen ();
	clearActions ();
	Updates = NULL;

	// Set the desirable action functions

	ScRefresh       = Refresh;
	ScUp            = Up;
	ScDown          = Down;
	ScLeft          = Left;
	ScRight         = Right;
	ScNext          = Next;
	ScPrev          = Prev;
	ScHelp          = Help;
	ScQuit          = Quit;
	ScExit          = Exit;
	ScOther         = Other;

	if (setjmp (done)) {

		// Exit to the caller
		return;
	}

	// The screen layout:
	// LUC -- The current owner (selectable), RUC -- "Hit '?' for help"
	// Second line: empty
	// Third + next: object list
	// Last-1: empty
	// LLC -- Current selection or command input buffer,
	// RLC -- Last error message

	setCurrentWindow (0, 2, NCols, SHeight = NRows-4);

	// Set the owner, top, and bottom
	Owner = Tree;
	// Determine the layout of items
	computeItemSize ();
	computeLayoutF (Owner->ChList);
	Selected = Owner;
	outLUC (Owner->dName ());
	outRUC ("Hit '?' for help");
	outLLC (Owner->dName ());
	dispScreen ();
	processUserInput ();
}
