/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ---------------------------------------------------------- */
/* This file contains functions presenting and processing the */
/* initial startup menu list                                  */
/* ---------------------------------------------------------- */

#include        <setjmp.h>
#include        "types.h"
#include        "var.h"

static  jmp_buf done;           // The return context
static  SmDesc  *Selected,      // The selected smurph
		*Top,           // The top smurph on the screen
		*Bottom;        // The bottom smurph on the screen

static  int     SHeight,        // The number of usable rows on the screen
		SelIndex,       // The window-relative row of selected smurph
		SelNumber;      // The number of the selected smurph

static  void    dispScreen () {

/* ------------------------------ */
/* Displays smurphs on the screen */
/* ------------------------------ */

	SmDesc  *sm;
	int     cnt, row, scn;

	// Locate the top smurph
	for (cnt = 0, sm = Smurphs; sm != Top; sm = sm->next, cnt++);

	SelNumber = NONE;
	scn = cnt;
	row = 1;

	while (1) {

		moveCursor (0, row);

		if (sm == Selected) {
			SelNumber = cnt;        // Locate the selected one
			SelIndex = row;
		}

		outText (DM_BRIGHT, form ("%3d. ", cnt));
		outText (rightAdjust (sm->HName, 16));
		outChar (' ');
		outText (rightAdjust (sm->DTime, 19));
		outChar (' ');
		outText (sm->PArgs);
		cnt++;
		row++;
		if (sm == Bottom) break;
		sm = sm->next;
	}

	if (SelNumber == NONE) {
		Selected = Top;
		SelNumber = scn;
		SelIndex = 1;
	}
	outLLC (form ("%1d", SelNumber));
	moveCursor (0, SelIndex);
}

static  void    dispHeader () {

/* ------------------------ */
/* Displays the header line */
/* ------------------------ */

	moveCursor (0, 0);      // This is window-relative
	outText (DM_BRIGHT,
		" No.  Running on host   Started date/time     Called as");
}

static  void    Refresh () {

/* -------------------- */
/* Refreshes the screen */
/* -------------------- */

	clearRLC ();                    // Cancel previous error message
	clearScreen ();
	refreshCorners ();
	dispHeader ();
	dispScreen ();
}

static  void    Up () {

/* --------------------------- */
/* Move the cursor one item up */
/* --------------------------- */

	clearRLC ();
	if (SelIndex < 2) {
		errMessage ("At the top");
		return;
	}
	outLLC (form ("%1d", --SelNumber));
	moveCursor (0, --SelIndex);
	Selected = Selected->prev;
}

static  void    Down () {

/* ----------------------------- */
/* Move the cursor one item down */
/* ----------------------------- */

	clearRLC ();
	if (SelIndex == NONE || SelIndex >= SHeight-2 ||
		Selected->next == NULL) {

		errMessage ("At the bottom");
		return;                 // No way
	}
	outLLC (form ("%1d", ++SelNumber));
	moveCursor (0, ++SelIndex);
	Selected = Selected->next;
}

static  void    Next () {

/* ----------------------- */
/* Move to the next screen */
/* ----------------------- */

	if (Bottom->next == NULL) {
		errMessage ("No more smurphs");
		return;
	}
	Top = Bottom->next;
	Refresh ();
}

static  void    Prev () {

/* --------------------------- */
/* Move to the previous screen */
/* --------------------------- */

	SmDesc  *sm;
	int     cnt;

	if (Top->prev == NULL) {
		errMessage ("This is the first screen");
		return;
	}
	for (sm = Top->prev, cnt = SHeight-2; sm -> prev != NULL && cnt > 0;
		sm = sm->prev, cnt--);
	Top = sm;
	Refresh ();
}

static  void    Quit () {

/* ------------------- */
/* Terminate execution */
/* ------------------- */

	clearScreen ();
	closeDisplay ();
	exit (0);
}

static  void    Exit () {

/* --------------------- */
/* Exit the current mode */
/* --------------------- */

	longjmp (done, 1);
}

static  void    Help () {

/* -------------------- */
/* Displays a help menu */
/* -------------------- */

	clearRLC ();                    // Cancel previous error message
	clearLLC ();
	clearScreen ();
	refreshCorners ();

	moveCursor (0, 0);
	outText (DM_REVERSE, "j   ");
	outText (" moves the cursor one line down");
	moveCursor (0, 1);
	outText (DM_REVERSE, "k   ");
	outText (" moves the cursor one line up");
	moveCursor (0, 2);
	outText (DM_REVERSE, "n   ");
	outText (" moves to the next screen");
	moveCursor (0, 3);
	outText (DM_REVERSE, "p   ");
	outText (" moves to the previous screen");
	moveCursor (0, 4);
	outText (DM_REVERSE, "r,^L");
	outText (" refresh the screen");
	moveCursor (0, 5);
	outText (DM_REVERSE, "q   ");
	outText (" exits the program");
	moveCursor (0, 6);
	outText (DM_REVERSE, "x   ");
	outText (" selects the current smurph and exits the menu");

	outLLC ("Hit any key to proceed");
	inChar ();
	clearLLC ();
	Refresh ();
}

SmDesc  *smurphMenu () {

/* ---------------------------------------------------------- */
/* Initialize  the menu processing, return the pointer to the */
/* description of the selected smurph                         */
/* ---------------------------------------------------------- */

	int     cnt;

	// Note: we assume that there is at least one active smurph; otherwise
	//       we wouldn't have gotten here in the first place

	clearActions ();

	// Set the desirable action functions

	ScRefresh       = Refresh;
	ScUp            = Up;
	ScDown          = Down;
	ScNext          = Next;
	ScPrev          = Prev;
	ScHelp          = Help;
	ScQuit          = Quit;
	ScExit          = Exit;

	if (setjmp (done)) {

		// Exit to the caller
		return (Selected);
	}

	// The screen layout:
	// LUC -- "Startup menu", RUC -- "Hit '?' for help"
	// Second line: empty
	// Third + next: smurph list
	// Last-1: empty
	// LLC -- "Current smurph", RLC -- "Last error message"

	setCurrentWindow (0, 2, NCols, SHeight = NRows-4);

	// Set up Top and Bottom
	for (cnt = SHeight-1, Bottom = Selected = Top = Smurphs;
		Bottom -> next != NULL && cnt > 1;
			Bottom = Bottom -> next, cnt--);
	outLUC ("Startup menu");
	outRUC ("Hit '?' for help");
	outLLC ("1");
	dispHeader ();
	dispScreen ();
	processUserInput ();
	return (Selected);      // Should never get here
}
