/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ---------------------------------------------------------- */
/* This file contains functions for initializing the internal */
/* template descriptions from system and user files           */
/* ---------------------------------------------------------- */

#include        <ctype.h>
#include        "fcntl.h"

#include        "types.h"
#include        "var.h"
#include        "tpath.h"

#ifndef EOF
#define EOF (-1)
#endif

class   RegDesc {

/* -------------------------------------------------------- */
/* A local data structure representing a region description */
/* -------------------------------------------------------- */

	friend  int     specialChar (int&, short*);
	friend  int     convertChar (short*);
	friend  class   Template;

	RegDesc *next, *prev;           // List pointers

	short   xs, xe;                 // Column bounds

	RegionField     *rfp;           // Field pointer

	public:

	RegDesc (short, short, RegionField*);
	~RegDesc ();
};

static  Template        *LastTemp;      // The last template built so far
static  int             TCounter,       // Template counter
			Column,         // Current line position
			Row,            // Current line number
			LineType,       // Template line type
			Width;          // Template width
	

static  Field           *LastLineStart, // Pointer to first field of this line
			*LastNotRepeat, // Last non-repeat field
			*LastField;     // Last processed field

static  RegDesc         *RegList = NULL;

static  char            *tefmsg;        // Template error message header

static  void            readTemplates ();
void                    fieldFlags ();

class   TFile {

	// This class describes a template file. Standard streams are not
	// used for efficiency reasons

	private:

	char    *buffer;        // Buffer
	int     filedesc;       // File descriptor
	char    *bufptr;        // Buffer pointer

	public:

	TFile (char*);
       ~TFile ();

	friend  int     getChar ();
	friend  int     peekChar ();
};

TFile::TFile (char *fname) {

	if ((filedesc = open (fname, O_RDONLY, 0)) < 0) {
		buffer = NULL;
		FError = YES;
	} else {        
		FError = NO;
		buffer = new char [TBUFSIZE+1];
		*buffer = '\0';
		bufptr = buffer;
	}
};

TFile::~TFile () {

	delete (buffer);
	close (filedesc);
};

TFile   *tf;

#ifndef ZZ_ATT
    inline
#endif
	int     getChar () {

/* ------------------------------------------ */
/* Read next character from the template file */
/* ------------------------------------------ */

	register char c;

	if ((c = *((tf->bufptr)++)) == '\0') {
		int     k;
		if ((k = read (tf->filedesc, tf->buffer, TBUFSIZE)) <= 0) {
			tf->bufptr--;
			return (EOF);
		} else {
			tf->buffer [k] = '\0';
			tf->bufptr = tf->buffer+1;
			Column++;
			return (*(tf->buffer));
		}
	}
	Column++;
	return (c);
};

#ifndef ZZ_ATT
    inline
#endif
	int     peekChar () {

/* ------------------------------------------------------ */
/* Look at the next character on input without reading it */
/* ------------------------------------------------------ */

	register char c;

	if ((c = *(tf->bufptr)) == '\0') {
		int     k;
		if ((k = read (tf->filedesc, tf->buffer, TBUFSIZE)) <= 0) {
			return (EOF);
		} else {
			tf->buffer [k] = '\0';
			return (*(tf->bufptr = tf->buffer));
		}
	}
	return (c);
};

void    tabt (char *msg) {

/* ------------------------ */
/* Processes template error */
/* ------------------------ */

	cerr << tefmsg << " template file: template #" << TCounter << " -- " <<
		msg << '\n';
	exit (1);
}

static  TFile *openUserTemplateFile () {

	return (new TFile ((utfname != NULL) ? utfname : UTEMPLATES));
}

static  TFile *openSystemTemplateFile () {

	if (stfname != NULL) return (new TFile (stfname));

	strcpy (CBuffer, ZZ_SOURCES);
	strcat (CBuffer, "/");
	strcat (CBuffer, STEMPLATES);

	return (new TFile (CBuffer));
}

RegDesc::RegDesc (short xst, short xen, RegionField *r) {

/* -------------------------------------- */
/* The constructor of a region descriptor */
/* -------------------------------------- */

	RegDesc *p, *q;

	xs = xst;
	xe = xen;
	rfp = r;

	// Put the new description onto sorted list

	if (RegList == NULL) {
		// The list is empty
		next = prev = NULL;
		RegList = this;
		return;
	}

	for (p = RegList; p != NULL; q = p, p = p->next) {
		if (p->xs >= xs) {
			if (p->xe >= xs)
				tabt ("overlapping regions");
			break;
		}
	}

	if (p == RegList) {
		// Add in front
		next = RegList;
		prev = NULL;
		RegList->prev = this;
		return;
	}

	q->next = this;
	prev = q;
	if ((next = p) != NULL) p->prev = this;
}

RegDesc::~RegDesc () {

/* --------------------------------------------------- */
/* Removes the region description from the sorted list */
/* --------------------------------------------------- */

	if (next != NULL) next->prev = prev;
	if (prev != NULL)
		prev->next = next;
	else
		RegList = next;
}

static  int     skipBlanks () {

/* ---------------------------------------------------------- */
/* Ignore   blanks  on  the  template  file,  stop  at  first */
/* non-blank without actually reading it                      */
/* ---------------------------------------------------------- */

	int     c;

	while ((c = peekChar ()) == ' ' || c == '\t') getChar ();
	return (c);
}

static  int     skipWhite () {

/* ------------------------ */
/* Skip blanks and newlines */
/* ------------------------ */

	int     c;

	do { c = skipBlanks (); } while (c == '\n' && getChar ());
	return (c);
}

static  void    skipToEol () {

/* -------------------------------------- */
/* Skip to the beginning of the next line */
/* -------------------------------------- */

	int     c;

	while ((c = getChar ()) != '\n')
		if (c == EOF) tabt ("incomplete template");
}

static  int     isLetter (int c) {

/* ---------------------------------------------------------- */
/* Checks  is  the  character is a legal first character of a */
/* name                                                       */
/* ---------------------------------------------------------- */

	if (isalpha (c)) return (YES);
	if (c == '_') return (YES);             // Underscore also legal
	return (NO);
}

static  int     isAlnum (int c) {

/* ---------------------------------------------------------- */
/* Checks  is  the  character  is  a legal character to occur */
/* within a name                                              */
/* ---------------------------------------------------------- */

	if (isalpha (c)) return (YES);
	if (c == '_') return (YES);             // Underscore also legal
	if (isdigit (c)) return (YES);
	return (NO);
}

static  char    *getName () {

/* ----------------------------------------- */
/* Read an identifier from the template file */
/* ----------------------------------------- */

	int     cnt;
	char    tmp [SANE], *r;

	if (!isLetter (peekChar ()))
		tabt ("type name starts with a non-letter");

	cnt = 1;
	tmp [0] = getChar ();

	while (isAlnum (peekChar ()))
		if (cnt >= SANE-1)
			tabt ("type name is too long");
		else
			tmp [cnt++] = getChar ();

	// Allocate memory for the string and return pointer to it

	tmp [cnt] = '\0';
	r = new char [cnt+1];
	strcpy (r, tmp);
	return (r);
}

static  long    getINumber () {

/* --------------------------------------------- */
/* Read an integer number from the template file */
/* --------------------------------------------- */

	int     c, pos;
	long    res;

	if ((c = peekChar ()) == '-') {
		pos = NO;
		getChar ();
	} else {
		pos = YES;
		if (c == '+') getChar ();
	}

	for (res = 0; isdigit (peekChar ()); res = res*10 - (getChar () - '0'));

	if (pos) res = -res;

	return (res);
}

static  double  getFNumber () {

/* --------------------------------------------------- */
/* Read a floating point number from the template file */
/* --------------------------------------------------- */

	int     c, pos, ex;
	double  res, fac;

	
	if ((c = peekChar ()) == '-') {
		pos = NO;
		getChar ();
	} else {
		pos = YES;
		if (c == '+') getChar ();
	}

	// The integer part
	for (res = 0.0; isdigit (c = peekChar ());
		res = res*10.0 + (double)(getChar () - '0'));

	if (c == '.') {
		// The fractional part
		getChar ();
		for (fac = 0.1; isdigit (c = peekChar ());
			res = res + fac * (double)(getChar () - '0'),
				fac *= 0.1);
	}

	if (c == 'e' || c == 'E') {
		// The exponent
		getChar ();
		for (ex = 0; isdigit (peekChar ());
			ex = ex * 10 + (getChar () - '0'));
		res *= pow (10.0, ex);
	}

	if (!pos)
		res = -res;

	return (res);
}

#ifdef  DEBUG

void    dumpTemplates () {

/* ------------------------------------ */
/* Dump all templates to the debug file */
/* ------------------------------------ */

	Template  *t;
	int       tc, i, fc;
	Field     *f, *g;

	FixedField  *fixf;
	RepeatField *repf;
	RegionField *regf;
	FilledField *filf;
	HeaderField *hdrf;

	for (tc = 1, t = TempList; t != NULL; tc++, t = t->next) {
		(*dbgf) << "Template #" << tc << "  " << t->typename << "  " <<
			t->mode << "  " << t->status << "  " << t->unbounded <<
				'\n';

		(*dbgf) << "  Clipping columns: (" << t->xclipd << ')';
		for (i = 0; i < t->nxclip; i++) (*dbgf) << ' ' << t->xclip [i];
		(*dbgf) << '\n';
		(*dbgf) << "  Clipping rows: (" << t->yclipd << ')';
		for (i = 0; i < t->nyclip; i++) (*dbgf) << ' ' << t->yclip [i];
		(*dbgf) << '\n';
		(*dbgf) << "  Field list:\n";

		for (fc = 1, f = t->fieldlist; f != NULL; fc++, f = f->next) {

		  (*dbgf) << "   Field type: ";
		  switch (f->type) {

		    case FT_FIXED:

			fixf = (FixedField*) f;
			(*dbgf) << "FIXED\n";
			(*dbgf) << "    DM: " << (int)(fixf->dmode) << 
				"  X: " << fixf->x << "  Y: " << fixf->y <<
					'\n';
			(*dbgf) << "    CN: " << fixf->contents << '\n';
			break;

		    case FT_REPEAT:

			repf = (RepeatField*) f;
			(*dbgf) << "REPEAT\n";

			(*dbgf) << "    From: ";
			for (i = 1, g = t->fieldlist; g != NULL; g = g->next,
				i++) if (g == repf->from) break;
			if (g == NULL)
				(*dbgf) << "unknown";
			else
				(*dbgf) << i;
			
			(*dbgf) << "  To: ";
			for (i = 1, g = t->fieldlist; g != NULL; g = g->next,
				i++) if (g == repf->to) break;
			if (g == NULL)
				(*dbgf) << "unknown";
			else
				(*dbgf) << i;

			(*dbgf) << "  Count: " << repf->count << '\n';
			break;
			
		    case FT_REGION:

			regf = (RegionField*) f;
			(*dbgf) << "REGION\n";
			(*dbgf) << "    DM: " << (int)(regf->dmode) << 
				"  X0: " << regf->x << "  Y0: " << regf->y <<
					"  X1: " << regf->x1 << "  Y1: " <<
						regf->y1 << '\n';
			(*dbgf) << "    PC: " << regf->pc << "  LC: " <<
				regf->lc << '\n';
			(*dbgf) << "    SC: " << regf->xo << ' ' << regf->xs <<
				' ' << regf->yo << ' ' << regf->ys << '\n';
			break;

		    case FT_FILLED:

			filf = (FilledField*) f;
			(*dbgf) << "FILLED\n";
			(*dbgf) << "    DM: " << (int)(filf->dmode) <<
				"  X: " << filf->x << "  Y: " << filf->y <<
					"  LT: " << filf->length << '\n';
			break;

		    case FT_HEADER:

			hdrf = (HeaderField*) f;
			(*dbgf) << "HEADER\n";
			(*dbgf) << "    DM: " << (int)(hdrf->dmode) <<
				"  X: " << hdrf->x << "  Y: " << hdrf->y <<
					"  LT: " << hdrf->length << '\n';
			break;

		    default:

			(*dbgf) << f->type << " (illegal)\n";

		  }
		}
	}
}
#endif

void    initTemplates    () {

/* ---------------------------------- */
/* Read templates from template files */
/* ---------------------------------- */

	LastTemp = NULL;                // Nothing done yet
	RegList = NULL;
	TCounter = 0;

	tf = openUserTemplateFile ();
	tefmsg = "User";
	if (!FError) {
		readTemplates ();
		delete (tf);
	}

	tf = openSystemTemplateFile ();
	tefmsg = "System";
	if (FError) {
		if (LastTemp == NULL)
			excptn ("no system template file");
	} else {

		TCounter = 0;
		readTemplates ();
		delete (tf);
	}

	if (LastTemp == NULL)
		excptn ("no templates defined");
#ifdef  DEBUG
	dumpTemplates ();
#endif
}

Field::Field () {

/* ------------------------- */
/* General field constructor */
/* ------------------------- */

	next = NULL;
	if (LastField == NULL)
		LastTemp->fieldlist = this;
	else
		LastField->next = this;

	LastField = this;
}

FixedField::FixedField (short xx, short yy) {

/* ----------------------- */
/* Fixed field constructor */
/* ----------------------- */

	type  = FT_FIXED;
	x     = xx;
	y     = yy;
	dmode = DM_NORMAL;              // Default display mode
	LastNotRepeat = this;
}

HeaderField::HeaderField (short xx, short yy) {

/* ----------------------- */
/* Fixed field constructor */
/* ----------------------- */

	type  = FT_HEADER;
	x     = xx;
	y     = yy;
	dmode = DM_REVERSE;             // Default header mode is 'reverse'
	LastNotRepeat = this;
}

FilledField::FilledField (short xx, short yy) {

/* ----------------------- */
/* Fixed field constructor */
/* ----------------------- */

	type  = FT_FILLED;
	x     = xx;
	y     = yy;
	dmode = DM_NORMAL;              // Default display mode
	LastNotRepeat = this;
}

RegionField::RegionField (short xx, short yy, short xx1) {

/* ------------------------ */
/* Region field constructor */
/* ------------------------ */

	type  = FT_REGION;
	x     = xx;
	y     = yy;
	x1    = xx1;
	y1    = -1;             // Unknown at the beginning
	dmode = DM_NORMAL;
	pc    = '*';            // Defaults
	lc    = '*';
	xo = xs = yo = ys = 0.0;
	LastNotRepeat = this;
}

RepeatField::RepeatField (int cnt) {

/* ------------------------ */
/* Repeat field constructor */
/* ------------------------ */

	type  = FT_REPEAT;
	from  = LastLineStart;
	to    = LastNotRepeat;
	count = cnt;
}

static  void    readTemplates () {

/* ---------------------------------------------------------- */
/* Read  templates  from  'tf'  and  add  them  to the global */
/* template list                                              */
/* ---------------------------------------------------------- */

	while (1) {
		TCounter++;
		if (skipWhite () == EOF) return;        // All done
		new Template;
	}
}

Template::Template () {

/* ------------------------------------------- */
/* Reads a new template from the template file */
/* ------------------------------------------- */

	int       c, NClip, More, LRow, i, j;
	short     Clip [SANE], Escape [SANE];

	next = NULL;
	unbounded = NO;
	if (TempList == NULL)
		TempList = this;
	else
		LastTemp->next = this;
	LastTemp = this;

	// Decode type name
	typename = getName ();

	if (skipBlanks () == EOF)
		tabt ("incomplete template header");

	if ((mode = (int)getINumber ()) < 0)
		tabt ("negative exposure mode");

	if (skipBlanks () == EOF)
		tabt ("incomplete template header");

	if (peekChar () == '*') {
		getChar ();
		if (peekChar () == '*')
			status = TS_FLEXIBLE;
		else
			status = TS_RELATIVE;
	} else
		status = TS_ABSOLUTE;

	// Skip everything to the first line starting with 'b' or 'B'

	while ((c = peekChar ()) != EOF) {
		if (c == '\n') {
			getChar ();
			Column = 0;
			continue;
		}
		if ((c == 'b' || c == 'B') && Column == 0) break;
		getChar ();
	}

	if (c == EOF)
		tabt ("EOF encountered before template started");

	// Here we have the first line of the template

	xclipd = yclipd = -1;   // Undefined yet

	getChar ();                     // Swallow the 'b'
	Column = 0;                     // Start counting from here
	NClip = 0;

	while ((c = peekChar ()) != 'b' && c != 'B' && c != '\n') {
		if (c == '|' || c == '+') {
			// Horizontal clipper
			if (NClip >= SANE)
				tabt ("too many clipping columns");
			Clip [NClip++] = Column;
			if (c == '+') {
				if (xclipd != -1)
			tabt ("default x-clipping defined more than once");
				xclipd = Column;
			}
		}
		getChar ();
	}

	Width = Column;

	if (xclipd == -1)               // Default clipping undefined
		xclipd = Width-1;

	if (NClip) {
		if (Clip [NClip-1] < Width-1) {
			if (NClip >= SANE)
				tabt ("too many clipping columns");
			Clip [NClip++] = Width-1;
		}

		xclip = new short [NClip];
		for (c = 0; c < NClip; c++) xclip [c] = Clip [c];
	} else
		xclip = NULL;

	nxclip = NClip;

	LRow = Row = 0;
	NClip = 0;
	More = YES;
	LastLineStart = LastField = LastNotRepeat = NULL;

	// Clear the escape array
	for (c = 0; c < Width; c++) Escape [c] = 0;

	while (More) {

	  skipToEol ();                 // Move to the beginning of next line

	  c = getChar ();
	  Column = 0;

	  switch (c) {

	    case 'e':                   // The end of the template
	    case 'E':

		More = NO;
		break;

	    case 'x':                   // The exceptions line
	    case 'X':

		LineType = LT_ESC;
		if (Row == -1)
			tabt ("'**' followed by 'exceptions' line");
		while ((c = peekChar ()) != '\n' && c != EOF) {
			if (Column < Width) {
				if (c == '|')
					Escape [Column] = ES_NSP;
				else if (c == '+')
					Escape [Column] = ES_FLD;
			}
			getChar ();
		}

		if (c == EOF)
			tabt ("Escape line incomplete");

		// Start for a new line with preset escape array
		continue;

	    case '|':                   // A regular field line

		LineType = LT_LAY;
		if (Row == -1)
			tabt ("'**' followed by layout line");
		layoutLine (Width, Escape);
		if (Row != -1) Row++;
		break;

	    case '*':                   // Replication

		LineType = LT_REP;
		if (Row == -1)
			tabt ("'**' followed by another replication line");
		if (Row == 0 || LastLineStart == NULL)
			tabt ("nothing to duplicate -- '*' illegal");

		if (RegList != NULL)
			tabt ("'*' illegal within region");

		if (peekChar () == '*') {
			if (yclipd == -1) yclipd = Row - 1;
			c = 0;          // Count undefined
			LRow = Row;
			Row = -1;       // Nondeterministic height
			unbounded = YES;
		} else {
			if ((c = (int)getINumber ()) < 0)
				tabt ("illegal repetition count");
			if (c == 0) c = 1;
			Row += c;
		}

		new RepeatField (c);
	  }

	  if (! More) break;            // After 'e' line

	  // Skip to '|' or EOL, whatever comes first

	  while ((c = peekChar ()) != '|' && c != '\n') {
		if (c == EOF)
			tabt ("incomplete template line");
		getChar ();
	  }

	  if (c == '|') {               // Possible attributes to follow
		getChar ();
		if ((c = peekChar ()) == '-' || c == '+') {
			// Clipping
			if (Row == -1)
			       tabt ("illegal cliping with undefined height");
			if (NClip >= SANE)
			       tabt ("too many clipping rows");
			Clip [NClip++] = Row-1;
			if (c == '+') {
				if (yclipd != -1)
			tabt ("default y-clipping defined more than once");
				yclipd = Row-1;
			}
			getChar ();
		}

		// Process field flags
		if (LineType == LT_LAY) fieldFlags ();
	  }

	  if (LineType == LT_LAY)
		// Clear the escape array
		for (c = 0; c < Width; c++) Escape [c] = 0;

	}       // End while (More)

	while ((c = getChar ()) != '\n') if (c == EOF) break;

	if (RegList != NULL)
		tabt ("incomplete region");

	// Vertical clipping

	if (yclipd == -1) {
		if (Row == -1)
			yclipd = LRow-1;
		else
			yclipd = Row-1;
	}

	if ((nyclip = NClip)) {
		for (i = 0; i < nyclip; i++)
			if (yclipd == Clip [i]) break;
		if (i >= nyclip) {
			if (NClip >= SANE)
			       tabt ("too many clipping rows");
			NClip++;
			for (i = 0; i < nyclip && Clip [i] < yclipd; i++);
			for (j = NClip-1; j > i; j--) Clip [j] = Clip [j-1];
			Clip [i] = yclipd;
			nyclip = NClip;
		}
		yclip = new short [nyclip];
		for (c = 0; c < nyclip; c++) yclip [c] = Clip [c];
	} else
		yclip = NULL;
}

int     specialChar (int &c, short *Escape) {

/* ---------------------------------------------------------- */
/* Checks  whether  the  next character is a special one. All */
/* characters within a region are special (blanks).           */
/* ---------------------------------------------------------- */

	RegDesc *r;

	if (Column >= Width) return (YES);

	// Check if within a region
	for (r = RegList; r != NULL; r = r->next) {
		// If you are afraid that this operation is too costly,
		// please remember that regions are rare and the 
		// likelihood of having more than one region open at
		// a time is practically zero
		if (r->xs > Column) break;
		if (r->xs == Column && peekChar () == '@' &&
			Escape [Column] != ES_NSP) {
			// Region termination
			c = '@';
			return (YES);
		}
		if (r->xe > Column) {
			c = ' ';
			return (YES);
		} else if (r->xe == Column && peekChar () == '@' &&
			Escape [Column] != ES_NSP) {
			c = '@';
			return (YES);
		}
	}

	c = peekChar ();
	if (Escape [Column] == ES_NSP) return (NO);

	if (c == '#' || c == '%' || c == '&' || c == '@' || c == ' ' ||
		c == '\n') return (YES);
	return (NO);
}

int     convertChar (short *Escape) {

/* ---------------------------------------------------------- */
/* Convert  a  character  to a normal printable form. Returns */
/* blanks if within a region.                                 */
/* ---------------------------------------------------------- */

	int     c;
	RegDesc *r;

	// Check if within a region

	for (r = RegList; r != NULL; r = r->next) {
		// Just relax
		if (r->xs > Column) break;
		if (r->xs == Column && peekChar () == '@' &&
			Escape [Column] != ES_NSP) {
			// Region termination
			return (getChar ());
		}
		if (r->xe > Column) {
			getChar ();
			return (' ');
		} else if (r->xe == Column && peekChar () == '@' &&
			Escape [Column] != ES_NSP) {
			// Termination
			return (getChar ());
		}
	}

	if ((c = getChar ()) == '~') // This stands for a honorary blank
		return (' ');
	else
		return (c);
}

void    Template::layoutLine (int wid, short *E) {

/* ----------------------- */
/* Process one layout line */
/* ----------------------- */

	FixedField      *FixF;
	FilledField     *FilF;
	RegionField     *RegF;
	HeaderField     *HdrF;

	int             FirstInLine = YES,
			c, i;

	char            Txt [SANE];

	while (1) {     // Loop for fields

		if (!specialChar (c, E)) {
			// The beginning of a regular 'as is' field
			FixF = new FixedField (Column, Row);
			if (FirstInLine) {
				LastLineStart = FixF;
				FirstInLine = NO;
			}

			Txt [0] = convertChar (E);
			for (i = 1; !specialChar(c, E) && E[Column] != ES_FLD;
				i++)
					Txt [i] = convertChar (E);

			Txt [i] = '\0';
			FixF->contents = new char [i+1];
			strcpy (FixF->contents, Txt);
			if (Column > wid)
				tabt ("fixed field outside window bounds");
			continue;
		}

		// Here we have the case of a special character

		if (Column >= wid) {
			// We have reached the end of line
			if (peekChar () != '|')
			     tabt ("layout line does not terminate with '|'");
			return;
		}

		if (c == '\n')
			tabt ("layout line ends prematurely");

		if (c == ' ') {         // Skip blanks
			while (1) {
				if (Column >= wid) break;
				specialChar (c, E);
				if (c == ' ')
					convertChar (E);
				else
					break;
			}
			continue;
		}

		if (c == '#') {         // Header field
			HdrF = new HeaderField (Column, Row);
			if (FirstInLine) {
				LastLineStart = HdrF;
				FirstInLine = NO;
			}
			HdrF->length = 1;
			convertChar (E);
			while (specialChar (c, E) && c == '#' &&
				E[Column] != ES_FLD && Column < wid) {

				convertChar (E);
				HdrF->length ++;
			}
			if (Column > wid)
				tabt ("header field outside window bounds");
			continue;
		}

		if (c == '%' || c == '&') {
			// Filled field
			int     sc;

			sc = c;         // The starting character
			FilF = new FilledField (Column, Row);
			FilF->just = (c == '%') ? RIGHT : LEFT;
			if (FirstInLine) {
				LastLineStart = FilF;
				FirstInLine = NO;
			}
			FilF->length = 1;
			convertChar (E);
			while (specialChar (c, E) && c == sc &&
				E[Column] != ES_FLD && Column < wid) {

				convertChar (E);
				FilF->length ++;
			}
			if (Column > wid)
				tabt ("filled field outside window bounds");
			continue;
		}

		if (c == '@') {         // Region

			int         x;
			RegDesc     *rd;

			x = Column;
			// Look for a non-special matching '@'
			convertChar (E);
			while (!specialChar (c, E) || c != '@') {
				if (Column >= wid)
					tabt ("unmatched region");
				else
					convertChar (E);
			}


			if (RegList != NULL) {
				// Check if it is not a closing frame
				c = NO;
				for (rd = RegList; rd != NULL; rd = rd->next)
					if (rd->xs == x) {
						if (rd->xe != Column)
						       tabt ("tilted region");
						rd->rfp->y1 = Row;
						delete (rd);
						c = YES;
						break;
					}

				if (c) {
					convertChar (E);
					continue;
				}
			}

			RegF = new RegionField (x, Row, Column);
			new RegDesc (x, Column, RegF);
			if (FirstInLine) {
				LastLineStart = RegF;
				FirstInLine = NO;
			}
			convertChar (E);
			continue;
		}

		tabt ("internal error -- illegal special character");
	}
}

static  float   xoo, xss, yoo, yss;     // Globals to return attribute values
static  int     ddmm;                   // Display mode
static  char    acc, bcc;               // Line/point display characters
static  int     mdset, pcset,           // Flags to detect duplicate setting
		lcset, scset;

int     static  getAttr () {

/* ---------------------------------------------------------- */
/* Decodes  one  attribute, stores it in the above variables, */
/* and returns the attribute type                             */
/* ---------------------------------------------------------- */

	int     c;

	while (1) {
		if ((c = skipBlanks ()) == ',') {
			getChar ();
			continue;
		}
		if (c == '\n') return (AT_END);
		break;
	}

	if (c == 'n' || c == 'N') {
		getChar ();
		ddmm = DM_NORMAL;
		return (AT_DMO);
	}

	if (c == 'r' || c == 'R') {
		getChar ();
		ddmm = DM_REVERSE;
		return (AT_DMO);
	}

	if (c == 'h' || c == 'H') {
		getChar ();
		ddmm = DM_BRIGHT;
		return (AT_DMO);
	}

	if (c == 'b' || c == 'B') {
		getChar ();
		ddmm = DM_BLINKING;
		return (AT_DMO);
	}

	if (c == '\'') {
		// Expect region display character(s)
		getChar ();
		if ((c = peekChar ()) == '\'' || c == '\n')
			tabt ("illegal display character specification");
		acc = c;
		getChar ();
		if ((c = peekChar ()) == '\'') {
			getChar ();
			return (AT_DCH);
		}
		if (c == '\n')
			tabt ("illegal display character specification");
		bcc = c;
		getChar ();
		if (peekChar () != '\'')
			tabt ("illegal display character specification");
		getChar ();
		return (AT_DCS);
	}

	if (isdigit (c) || c == '.' || c == '+' || c == '-') {
		// Scaling specification
		
		xoo = getFNumber ();
		c = skipBlanks ();
		if (c == ',') {
			getChar ();
			c = skipBlanks ();
		}
		xss = getFNumber ();
		c = skipBlanks ();
		if (c == ',') {
			getChar ();
			c = skipBlanks ();
		}
		yoo = getFNumber ();
		c = skipBlanks ();
		if (c == ',') {
			getChar ();
			c = skipBlanks ();
		}
		yss = getFNumber ();
		c = skipBlanks ();
		if (c == ',') {
			getChar ();
			c = skipBlanks ();
		}

		if (xss == 0.0 || yss == 0.0)
			tabt ("illegal scaling attribute");
		return (AT_SCA);
	}

	return (AT_END);
}

/* ------------------------------- */
/* Simple functions for fieldFlags */
/* ------------------------------- */

static  void    setDMode (Field *f) {

	if (f == NULL)
		tabt ("superfluous field attribute(s)");

	if (mdset)
		tabt ("display mode set twice for the same field");

	mdset = YES;
	((FilledField*)f)->dmode = ddmm;
}

static  void    setDChar (Field *f) {

	RegionField     *rf;

	if (f == NULL)
		tabt ("superfluous field attribute(s)");

	if (f->type != FT_REGION)
		tabt ("attempt to set display char for a non-region field");

	rf = (RegionField*) f;

	if (!pcset) {
		rf->lc = rf->pc = acc;
		pcset = YES;
		return;
	}

	if (lcset)
		tabt ("attempt to set display char twice");

	lcset = YES;
	rf->lc = acc;
}

static  void    setDChrs (Field *f) {

	RegionField     *rf;

	if (f == NULL)
		tabt ("superfluous field attribute(s)");

	if (f->type != FT_REGION)
		tabt ("attempt to set display chars for a non-region field");

	rf = (RegionField*) f;

	if (pcset)
		tabt ("attempt to set display chars twice");

	pcset = lcset = YES;
	rf->pc = acc;
	rf->lc = bcc;
}

static  void    setDScal (Field *f) {

	RegionField     *rf;

	if (f == NULL)
		tabt ("superfluous field attribute(s)");

	if (f->type != FT_REGION)
		tabt ("attempt to set display scale for a non-region field");

	rf = (RegionField*) f;

	if (scset)
		tabt ("attempt to set display scale twice");

	scset = YES;
	rf->xo = xoo;
	rf->xs = xss;
	rf->yo = yoo;
	rf->ys = yss;
}

void    fieldFlags () {

/* -------------------------------------------- */
/* Process field flags for the last layout line */
/* -------------------------------------------- */

	Field           *f;
	int             More;

	f = LastLineStart;      // The first field of the last line

	while (1) {

		mdset = pcset = lcset = scset = NO;

		if (skipBlanks () == '(') {
			// A group
			getChar ();
			for (More = YES; More;) {
				switch (getAttr ()) {

					case AT_DMO:
						setDMode (f);
						break;
					case AT_DCH:
						setDChar (f);
						break;
					case AT_DCS:
						setDChrs (f);
						break;
					case AT_SCA:
						setDScal (f);
						break;
					case AT_END:
						if (getChar () != ')')
			tabt ("missing ')' in attribute specification");
						More = NO;
				}
			}
		} else {

			switch (getAttr ()) {
				case AT_DMO:
					setDMode (f);
					break;
				case AT_DCH:
					setDChar (f);
					break;
				case AT_DCS:
					setDChrs (f);
					break;
				case AT_SCA:
					setDScal (f);
					break;
				case AT_END:
					if (peekChar () != '\n')
				tabt ("illegal attribute specification");
					else
						return;
			}
		}

		if (f != NULL) f = f->next;
		if (f == NULL) {
			if (skipBlanks () != '\n')
				tabt ("too many attribute specifications");
			return;
		}
	}
}
