/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */

/* ------------------------------------------ */
/* This is plain C. Should be compiled by cc. */
/* ------------------------------------------ */

#include <fcntl.h>

int zz_fopenf (fn, mo) 

    char *fn, *mo;

{
	if (strcmp (mo, "r") == 0)
		/* For reading */
		return (open (fn, O_RDONLY, 0));
	else if (strcmp (mo, "a") == 0)
		return (open (fn, O_WRONLY + O_CREAT + O_APPEND, 0666));
	else
		return (open (fn, O_WRONLY + O_CREAT + O_TRUNC, 0666));
};

/* ---------------------------------------------------------------------- */
/* Memory allocator/deallocator. This is plain C -- to be compiled by cc  */
/* ---------------------------------------------------------------------- */

/* ---------------------------------------------------------------------- */
/* Note:  this  memory  allocator/deallocator  has  been adapted from the */
/* LANSF allocator/deallocator. It facilitates allocation of small frames */
/* of several (standard) sizes which seems to be quite typical in SMURPH. */
/* ---------------------------------------------------------------------- */

#define	MATSIZE  256
#define	VIPSIZE	 (MATSIZE * 8 * 2)

#define	excptn(a)    excptn__FPc (a)	/* This is a C++ function */

/* ------------------------------------------- */
/* Direct pointers to free lists size 8 - 2040 */
/* ------------------------------------------- */

static	char	*piece_list [MATSIZE] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

static	char	*last_free = (char*)0,	/* FWA of contiguous free storage */
		*last_limit = (char*)0,	/* LWA+1 of free storage */
            	*bfree = (char*)0;	/* Tree off "big" pieces */

extern	long	MemoryUsed;		/* Memory balance (used by smurph) */

static	unsigned	noctets,	/* Requested size / 8 - 1 */
			asize;		/* Actual size of the piece */

#define	itag(p)	(*((unsigned int*)(p)))
#define	ptag(p)	(*((char**)(p)))
#define intp(p) ((unsigned int)(p))

static	char	*get_big (), *reclaim ();

#ifdef	DEBUG
outt (t)

char *t;

{
	write (2, t, strlen (t));
}

excptn__FPc (t)

char *t;

{
	outt (t); outt ("\n"); exit (11);
}
#else
#define	outt(a)
#endif

char	*sbrk ();

char	*malloc (size)

/* ----------------------------- */
/* Our private version of malloc */
/* ----------------------------- */

unsigned	long	size;

{
register	char	*optr;

	asize = ((noctets = ((size + 3) >> 3) & 03777777777) + 1) << 3;
	if (noctets >= MATSIZE) {
		if (noctets > 0x0fffffff)
			excptn ("malloc: illegal memory request");
		/* This is a "big" piece */
		return (get_big ());
	}

	/* Here we have a small piece */
	if ((optr = piece_list [noctets]) != 0) {
		/* Direct hit: take a free piece from the list */
		piece_list [noctets] = ptag (optr);
		MemoryUsed += asize;
		return (optr);
	}

	if ((unsigned) last_limit  -  (unsigned) last_free < asize) {
		/* Not enough room: request new page */
		if ((optr = sbrk (VIPSIZE)) == (char*) (-1)) {
			/* Running out of space - be desperate */
			if ((optr = reclaim ()) == 0) {
				excptn ("malloc: out of space");
			}
			return (optr);
		}
		if (optr == last_limit) {	/* A contiguous piece */
			last_limit = (char*)((int) last_limit + VIPSIZE);
		} else {
			outt ("Non contiguous piece\n");
			/* Make sure it starts at a doubleword boundary */
			last_free = optr+4;
			last_limit = (char*)((int) optr + VIPSIZE);
		}
	}

	optr = last_free+4;
	itag (last_free) = 0x30000000 | noctets;
	last_free += asize;
	MemoryUsed += asize;
	return (optr);
}

static	char	*get_big () {

	char		*cl, *pl, *ql;
	unsigned int	pt;

	pt = 0x30000000 | noctets;
	for (cl = bfree, pl = ((char*)(&bfree)) - 4; cl; pl = cl,
	    cl = ptag (cl+4)) {
		if (itag (cl-4) >= pt) {
			/* Found */
			if (ql = ptag (cl)) {
				/* The list will stay nonempty */
				ptag (pl+4) = ql;
				ptag (ql+4) = ptag (cl+4);
			} else {
				/* The list becomes empty */
				ptag (pl+4) = ptag (cl+4);
			}
			MemoryUsed += ((itag (cl-4) & 0x0fffffff) + 1) << 3;
			return (cl);
		}
	}

	/* When we are here, no frame of the requested size is available */

	while ((unsigned) last_limit  -  (unsigned) last_free < asize) {
		/* Not enough room: request new page */
		if ((cl = sbrk (VIPSIZE)) == (char*) (-1)) {
			/* Out of space */
			excptn ("malloc: out of space");
		}
		if (cl == last_limit) {	/* A contiguous piece */
			last_limit = (char*)((int) last_limit + VIPSIZE);
		} else {
			outt ("Non contiguous piece\n");
			/* Make sure it starts at a doubleword boundary */
			last_free = cl+4;
			last_limit = (char*)((int) cl + VIPSIZE);
		}
	}

	cl = last_free+4;
	itag (last_free) = 0x30000000 | noctets;
	last_free += asize;
	MemoryUsed += asize;
	return (cl);
}

static	char	*reclaim () {

	char		*cl, *ql, **pl;
	unsigned int	pt, co, fu;

	/* We get here when a request for a small piece cannot be fulfilled */

	outt ("reclaim called\n");
	if (!bfree) {
		/* Try a bigger small frame */
		for (noctets++; noctets < MATSIZE; noctets++) {
			if (piece_list [noctets]) {
				ql = piece_list [noctets];
				piece_list [noctets] = ptag (ql);
				MemoryUsed += (noctets + 1) << 3;
				return (ql);
			}
		}
		excptn ("malloc: out of space");
	}
	/* Remove the first big piece from the list */
	cl = bfree;
	if (ql = ptag (cl)) {
		/* The list will stay nonempty */
		bfree = ql;
		ptag (bfree+4) = ptag (cl+4);
	} else {
		/* The list becomes empty */
		bfree = ptag (cl+4);
	}

	pt = itag (cl-4) & 0x0fffffff - noctets;

	ql = cl;
	itag (cl-4) = (fu = 0x30000000 | noctets);

	co = noctets + 1;
	cl += asize;
	pl = &(piece_list [noctets]);


	while (pt >= co) {
		itag (cl-4) = fu;
		ptag (cl) = *pl;
		*pl = cl;
		pt -= co;
		cl += asize;
	}

	MemoryUsed += asize;
	return (ql);
}
	
free (e)

/* ------------------ */
/* Memory deallocator */
/* ------------------ */

char		*e;

{

	register unsigned int 	f, ff;
	char	*cl, *pl, *ql;

	if ((f = itag (e-4)) & 0x20000000) {
		ff = f & 0x0fffffff;
		MemoryUsed -= (ff + 1) << 3;
		if (ff < MATSIZE) {
			/* Put back into a piece list */
			ptag (e) = piece_list [ff];
			piece_list [ff] = e;
			return;
		}
	
		/* Deallocate a big piece */
	
		for (cl = bfree, pl = ((char*)(&bfree)) - 4; cl; pl = cl,
		    cl = ptag (cl+4)) {
			if (itag (cl-4) >= f) {
				if (itag (cl-4) != f) break;
				/* Insert into an existing list */
				ptag (pl+4) = e;
				ptag (e+4) = ptag (cl+4);
				ptag (e) = cl;
				return;
			}
		}
	
		/* Create a new list */
	
		ptag (e+4) = cl;
		ptag (pl+4) = e;
		ptag (e) = 0;
	} else
		excptn ("free: corrupted frame");
}

char *realloc (ptr, size)

char *ptr;  unsigned int size;

/* ------------------------------------------------- */
/* This is just for completeness. Not very efficient */
/* ------------------------------------------------- */

{
	unsigned int f;
	char	 *nptr;

	f = itag (ptr-4);
	if ((f & 0xf0000000) != 0x30000000)
		excptn ("realloc: corrupted frame pointer");
	f = ((f & 0x0fffffff) + 1) << 3;

	if (size <= f) return (ptr);

	nptr = malloc (size);
	bcopy (ptr, nptr, f);
	free (ptr);
	return (nptr);
}
