/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */

/* ------------------------------------------------------------- */
/* SMURPH version of the C++ library (the minimum needed subset) */
/* ------------------------------------------------------------- */

#include "../version.h"

#include "lib.h"
#include <stdarg.h>

#define	NFMTBUF	4	// Number of buffers for form
#define	FMTBSIZ 256	// Size of buffer for form

extern "C" {
		int	zz_fopenf (char*, char*);
#ifndef	ZZ_ATT_SUN
		void	vsprintf (char*, const char*, ...);
#endif
};

	// Sorry, but ctype.h includes stdio.h, which I don't want
#define	isdigit(c)  ((c) >= '0' && (c) <= '9')
#define isspace(c)  ((c) == ' ' || (c) == '\n' || (c) == '\t' || (c) == '\f')

istream	cin = (0);
ostream	cout = (1);
ostream cerr = (2);

typedef	void(*EHTYPE)(char*);

static	EHTYPE	eh = NULL;

void	set_File_error_handler (EHTYPE p) {
	eh = p;
};

static	void aberr (char *t) {

	write (2, t, strlen (t));
	exit (111);
};

istream::istream (char *fn, char *mode) {

	char	tmp [256];

	if (strcmp (mode, "r")) {
		strcpy (tmp, "open failed for ");
		strcat (tmp, fn);
		strcat (tmp, ": illegal mode");
		if (eh != NULL) {
			(*eh) (tmp);
			return;
		} else {
			strcat (tmp, "\n");
			aberr (tmp);
		}
	}

	if ((fd = zz_fopenf (fn, mode)) < 0) {
		strcpy (tmp, "open failed for ");
		strcat (tmp, fn);
		if (eh != NULL) {
			(*eh) (tmp);
			return;
		} else {
			strcat (tmp, "\n");
			aberr (tmp);
		}
	}
	end = 0;
	ungp = 0;
};

void istream::close () { ::close (fd); };
void ostream::close () { flush (); ::close (fd); };
ostream::~ostream () { close (); };

void ostream::flush () {

	if (bp) {
		write (fd, buf, bp);
		bp = 0;
	}
};

istream::istream (int des) {

	fd = des;
	end = 0;
	ungp = 0;
};

ostream::ostream (char *fn, char *mode) {

	char	tmp [256];

	bp = 0;
	if (strcmp (mode, "w") && strcmp (mode, "a")) {
		strcpy (tmp, "open failed for ");
		strcat (tmp, fn);
		strcat (tmp, ": illegal mode");
		if (eh != NULL) {
			(*eh) (tmp);
			return;
		} else {
			strcat (tmp, "\n");
			aberr (tmp);
		}
	}

	if ((fd = zz_fopenf (fn, mode)) == NULL) {
		strcpy (tmp, "open failed for ");
		strcat (tmp, fn);
		if (eh != NULL) {
			(*eh) (tmp);
			return;
		} else {
			strcat (tmp, "\n");
			aberr (tmp);
		}
	}
};

ostream::ostream (int des) {

	fd = des;
	bp = 0;
};

void istream::get (char &c) {

	if (end) { c = ' '; return; }
	if (ungp) { c = ung [--ungp]; return; }
	if (read (fd, &c, 1) <= 0) {
		end = 1;
		c = ' ';
	}
};

int istream::readable () {

	char tmp;

	if (end) return (0);
	get (tmp);
	if (end) return (0);
	putback (tmp);
	return (1);
};

int istream::eof () {
	return (! readable ());
};

void istream::putback (char c) {
	if (ungp < 8) {
		ung [ungp++] = c;
		end = 0;
	}
};

int	istream::digg (char *b, int &rc) {

	char c;

	while (rc < 255) {
		if (end) return (-1);
		get (c);
		if (end) return (-1);
		if (!isdigit (c)) return (c);
		*b++ = c;
		rc++;
	}
};

istream &operator >> (istream &is, long &n) {

	char tmp [256], c;
	int  i, k;

	for (i = 0; ;) {
		if (is.end) {
			n = 0L;
			return (is);
		}
		is.get (c);
		if (!isspace (c)) break;
	}

	if (c == '-' || c == '+' || isdigit (c)) {
		tmp [i++] = c;
		if ((k = is.digg (&(tmp [i]), i)) > 0) is.putback (k);
		if (i == 1 && (tmp[0] == '+' || tmp[0] == '-')) {
			is.putback (tmp[0]);
			n = 0L;
			return (is);
		}
		tmp [i] = '\0';
		n = atol (tmp);
	} else {
		is.putback (c);
		n = 0L;
	}
	return (is);
};

istream &operator >> (istream &is, int &n) {

	long ln;

	is >> ln;
	n = (int) ln;
	return (is);
};

istream &operator >> (istream &is, double &n) {

	char tmp [256], c;
	int  i, k, ok;

	for (i = 0; ;) {
		if (is.end) {
			n = 0.0;
			return (is);
		}
		is.get (c);
		if (!isspace (c)) break;
	}

	i = 0;
	if (c == '-' || c == '+') {
		tmp [i++] = c;
		is.get (c);
		if (is.end) {
Cancel:
			while (i) is.putback (tmp [--i]);
			n = 0.0;
			return (is);
		}
	}

	// Sign part done

	ok = 0;
	if (isdigit (c)) {
		// The integer part
		tmp [i++] = c;
		ok = 1;
		k = is.digg (&(tmp[i]), i);
		if (k < 0) {
Done:
			if (!ok) goto Cancel;
			tmp [i] = '\0';
			n = atof (tmp);
			return (is);
		}
		c = k;
	}

	if (c == '.') {
		if (i >= 256) goto Done;
		tmp [i++] = c;
	} else if (c == 'e' || c == 'E') goto Exponent;
		else {
			is.putback (c);
			goto Done;
		}

	// The fraction

	if (is.end) goto Done;
	is.get (c);
	if (isdigit (c)) {
		if (i >= 256) {
			is.putback (c);
			goto Done;
		}
		ok = 1;
		tmp [i++] = c;
		k = is.digg (&(tmp[i]), i);
		if (k < 0) goto Done;
		c = k;
	}

	if (c == 'e' || c == 'E') {
Exponent:
		if (!ok) goto Cancel;
		if (i >= 256) {
			is.putback (c);
			goto Done;
		}
		tmp [i++] = c;
		if (is.end) goto Done;
		is.get (c);
		if (c == '-' || c == '+') {
			if (i >= 256) {
				is.putback (c);
				goto Done;
			}
			tmp [i++] = c;
			is.get (c);
		}
		if (!isdigit (c)) goto Done;
		if (i >= 256) {
			is.putback (c);
			goto Done;
		}
		tmp [i++] = c;
		k = is.digg (&(tmp[i]), i);
		if (k > 0) is.putback (k);
		goto Done;
	}
	is.putback (c);
	goto Done;
};

istream &operator >> (istream &is, float &n) {

	double d;

	is >> d;
	n = d;
	return (is);
};

istream &operator >> (istream &is, char &n) {

	is.get (n);
	return (is);
};

void ostream::put (char c) {

	if (bp >= OBUFSIZE) {
		write (fd, buf, OBUFSIZE);
		bp = 0;
	}
	buf [bp++] = c;
/* ==== removed
	if (c == '\n') {
		write (fd, buf, bp);
		bp = 0;
	}
   ==== */
};

ostream &operator << (ostream &os, long n) {

	char tmp [256]; int i;

	sprintf (tmp, "%1d", n);
        for (i = 0, n = strlen (tmp); i < n; i++) os.put (tmp [i]);
	return os;
};

ostream &operator << (ostream &os, int n) {

	char tmp [256]; int i;

	sprintf (tmp, "%1d", n);
        for (i = 0, n = strlen (tmp); i < n; i++) os.put (tmp [i]);
	return os;
};

ostream &operator << (ostream &os, char n) {

	os.put (n);
	return os;
};

ostream &operator << (ostream &os, double n) {

	char tmp [256]; int i, j;

	sprintf (tmp, "%g", n);
        for (i = 0, j = strlen (tmp); i < j; i++) os.put (tmp [i]);
	return os;
};

ostream &operator << (ostream &os, char *n) {

	int i, j;

        for (i = 0, j = strlen (n); i < j; i++) os.put (n [i]);
	return os;
};

static	char fmtbuf [NFMTBUF][FMTBSIZ];
static	int  cbi = 0;

char *form (const char *f, ...) {

	va_list pmts;
	char	*tar;

	va_start (pmts, f);
	if (cbi >= NFMTBUF) cbi = 0;
	tar = fmtbuf [cbi++];
	vsprintf (tar, f, pmts);
	return (tar);
};
