/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ------------------------------------------------------- */
/* Global functions related to AIs and displayable objects */
/* ------------------------------------------------------- */

#include        "system.h"

#if     ZZ_SMURPHDEBUG == 2
static  void    zz_dumpTree (ZZ_Object *o) {

/* -------------------- */
/* Dumps ownership tree */
/* -------------------- */

	ZZ_Object  *clist, *osent;

	switch (o->Class) {

		case    OBJ_station:

			clist = ((Station*)o)->ChList;
			osent = (ZZ_Object*) (&((Station*)o)->ChList);
			break;

		case    AIC_traffic:

			clist = ((Traffic*)o)->ChList;
			osent = (ZZ_Object*) (&((Traffic*)o)->ChList);
			break;

		case    AIC_process:

			clist = ((Process*)o)->ChList;
			osent = (ZZ_Object*) (&((Process*)o)->ChList);
			break;

		default:

			return;         // Nothing to do
	}

	if (clist == NULL) return;      // Same thing

	// Go to the end of the object list

	while (clist->next != NULL) clist = clist->next;

	// Now go backwards

	do {
		debug ("---------------------------");
		debug (form ("Parent: %s", o->getSName ()));
		debug (form ("Object: %s", clist->getSName ()));
		debug (form ("BasNam: %s", clist->getBName ()));
		debug (form ("NikNam: %s", (clist->zz_nickname == NULL) ?
			"NONE" : clist->zz_nickname));
		zz_dumpTree (clist);    // Recursively for the child

	} while ((clist = clist->prev) != osent);
}
#endif

void    zz_adjust_ownership () {

/* ---------------------------------------------------------- */
/* Reorganizes  the  ownership  tree structure after Root has */
/* completed initialization                                   */
/* ---------------------------------------------------------- */

	ZZ_Object  *o, *so, *ox;
	int        i;

	Assert (!zz_flg_otreeAdjusted,
		"Internal error: ownership tree adjusted twice");
	zz_flg_otreeAdjusted = YES;

	// Move all stations to System

	Assert (System->ChList == NULL,
	"Adjusting ownership tree: internal error -- System ChList not empty");

	queue_head ((ZZ_Object*)Kernel, System->ChList, ZZ_Object);

	for (i = 0; i < NStations; i++) {
		o = idToStation (i);
		queue_out (o);
		queue_head (o, System->ChList, ZZ_Object);
		// Note that all objects are linked at their owners in the
		// reverse order of creation
	}

	// Now, move all links to System

	for (i = 0; i < NLinks; i++) {

		o = idToLink (i);
		queue_out (o);
		queue_head (o, System->ChList, ZZ_Object);
	}

	// Same with traffic patterns

	for (i = 0; i < NTraffics; i++) {

		o = idToTraffic (i);
		queue_out (o);
		queue_head (o, System->ChList, ZZ_Object);
	}

	// All fixed standard AI's

	queue_head ((ZZ_Object*)Client, System->ChList, ZZ_Object);
	queue_head ((ZZ_Object*)Timer, System->ChList, ZZ_Object);
	queue_head ((ZZ_Object*)Signal, System->ChList, ZZ_Object);
	queue_head ((ZZ_Object*)Queue, System->ChList, ZZ_Object);

	// Assign ports to their stations

	for (i = 0; i < NStations; i++) {

		Station *st;

		st = idToStation (i);
		for (o = st->Ports; o != NULL; o = ((Port*)o)->nextp) {
			queue_out (o);
			queue_head (o, st->ChList, ZZ_Object);
		}
	}

	// Assign level 0 processes to their stations as well

	if ((so = ZZ_Main->ChList) != NULL) {
		// The process list nonempty (normally it should be)
		for (o = so; o->next != NULL; o = o->next);
		// Traverse the Kernel list backwards -- to retain the
		// creation order
		while (1) {

			if (o->Class != AIC_process ||
				((Process*)o)->Owner == System) {

				// System processes remain at Kernel
				if (o == so) break;
				o = o->prev;
				continue;
			}

			assert (((Process*)o)->Owner != NULL,
			   "System: a process without owner -- internal error");

			ox = o->prev;

			queue_out (o);
			queue_head (o, ((Process*)o)->Owner->ChList, ZZ_Object);
			if (o == so) break;
			o = ox;
		}
	}

	// Move RVariables at stations to the end of the children list (actually
	// they have to be moved to the front; the list is reversed!)

	for (i = 0; i < NStations; i++) {

		Station *st;

		st = idToStation (i);

		for (o=st->ChList, ox=(ZZ_Object*)(&(st->ChList)); o!=NULL;) {

			if (o->Class != OBJ_rvariable) {
				o = o->next;
				continue;
			}

			so = o->next;

			queue_out (o);
			o->prev = ox;
			o->next = ox->next;
			if (o->next != NULL) o->next->prev = o;
			ox->next = o;
			ox = o;
			o = so;
		}
	}

	// Move User-defined displayable objects created at level-0 (belonging
	// to stations) to the end of the children list (actually
	// they have to be moved to the front; the list is reversed!)

	for (i = 0; i < NStations; i++) {

		Station *st;

		st = idToStation (i);

		for (o=st->ChList, ox=(ZZ_Object*)(&(st->ChList)); o!=NULL;) {

			if (o->Class != OBJ_eobject) {
				o = o->next;
				continue;
			}

			so = o->next;

			queue_out (o);
			o->prev = ox;
			o->next = ox->next;
			if (o->next != NULL) o->next->prev = o;
			ox->next = o;
			ox = o;
			o = so;
		}
	}

	// Sort the Kernel's queue

	ox = Kernel->ChList;
	Assert (ox != NULL,             // At least Root should be there
		"Adjusting ownership tree: Kernel's ChList is empty");
	ox->prev = (ZZ_Object*)(&ox);   // Make it consistent
	Kernel->ChList = NULL;

	// Find Root: it goes first

	for (o = ox; o != NULL && o != ZZ_Main; o = o->next);
	if (o == NULL)
	   excptn ("Adjusting ownership tree: Root missing from Kernel's list");

	queue_out (o);
	queue_head (o, Kernel->ChList, ZZ_Object);

	// Now for system processes

	for (o = ox; o != NULL; ) {
		if (o->Class != AIC_process || ((Process*)o)->Owner != System) {
			o = o->next;
			continue;
		}

		// A system process
		so = o->next;
		queue_out (o);
		queue_head (o, Kernel->ChList, ZZ_Object);
		o = so;
	}

	// Now for the Root's list

	for (o = ZZ_Main->ChList, ox = (ZZ_Object*)(&(ZZ_Main->ChList));
		o != NULL;) {

		// Observers first

		if (o->Class != AIC_observer) {
			o = o->next;
			continue;
		}

		so = o->next;

		queue_out (o);
		o->prev = ox;
		o->next = ox->next;
		if (o->next != NULL) o->next->prev = o;
		ox->next = o;
		ox = o;
		o = so;
	}

	for (o = ZZ_Main->ChList, ox = (ZZ_Object*)(&(ZZ_Main->ChList));
		o != NULL;) {

		// Then RVariables

		if (o->Class != OBJ_rvariable) {
			o = o->next;
			continue;
		}

		so = o->next;

		queue_out (o);
		o->prev = ox;
		o->next = ox->next;
		if (o->next != NULL) o->next->prev = o;
		ox->next = o;
		ox = o;
		o = so;
	}

	// Whatever just remains, it must be level-0 user-defined displayable
	// objects unasigned to any station; so they should stay right where
	// they are.
#if     ZZ_SMURPHDEBUG == 2
	zz_dumpTree (System);
#endif
}

void    ZZ_Object::zz_start () {

/* --------------- */
/* This is a dummy */
/* --------------- */

	excptn ("Object: cannot create directly objects of this class");
}

char *ZZ_Object::getNName () {

/* ----------------------------- */
/* Returns the object's nickname */
/* ----------------------------- */

	return (zz_nickname);
}

char *ZZ_Object::getOName () {

/* ----------------------------------- */
/* Returns the object's printable name */
/* ----------------------------------- */

	char    *s;

	if ((s = getNName ()) != NULL)
		return (s);
	else
		return (getSName());
}

char *ZZ_Object::getSName () {

/* ---------------------------------- */
/* Returns the object's standard name */
/* ---------------------------------- */
	
	if (Id == NONE) return (getTName ());

	if (Class == AIC_port)      // To be treated in a special way
		return (form ("%s %1d at %s", getTName (), Id & 0177777,
			idToStation ((Id >> 16) & 0177777)->getSName ()));
	else
		// Compose name and Id
		return (form ("%s %1d", getTName (), Id));
}

char *ZZ_Object::getBName () {

/* --------------------------------------- */
/* Returns the object's base standard name */
/* --------------------------------------- */
	
	switch (Class) {

		case AIC_timer:         // All these objects are not
		case AIC_queue:         // extensible by the user
		case AIC_signal:
		case AIC_client:
		case AIC_port:
		case OBJ_rvariable:

			return (getSName ());

		case AIC_link:

			return (form ("Link %1d", Id));

		case AIC_traffic:

			return (form ("Traffic %1d", Id));

		case AIC_process:

			return ((this == Kernel) ? getSName () :
				form ("Process %1d", Id));

		case AIC_observer:

			return (form ("Observer %1d", Id));

		case OBJ_station:

			return ((this == System) ? getSName () :
				form ("Station %1d", Id));

		case OBJ_eobject:

			return (form ("EObject %1d", Id));

		default:

			excptn (form ("getBName: illegal object class %1d",
				Class));
	}
}

void    EObject::zz_start () {

/* ---------------------------------------------------------- */
/* Nonstandard  constructor  for  user-definable  displayable */
/* objects                                                    */
/* ---------------------------------------------------------- */

	Id = sernum++;                  // Update numerical Id

	Class = OBJ_eobject;

	if (TheProcess == ZZ_Main && TheStation != NULL &&
		TheStation != System) {

		// Created at level 0 -- assign to the user station

		assert (!zz_flg_otreeAdjusted,
		 "EObject: cannot create from Root after protocol has started");

		zz_queue_head ((ZZ_Object*)this, TheStation->ChList, ZZ_Object);

	} else {

		// Not level 0 -- assign to the rightful owner, i.e. the
		// process that creates the object

#if     ZZ_OBS
		if (zz_observer_running) {
			// Observers are also allowed to create dynamic
			// displayable objects
			zz_queue_head (this, zz_current_observer->ChList,
				ZZ_Object);
			zz_DADD (zz_current_observer, this);
		} else
#endif
		{
			zz_queue_head (this, TheProcess->ChList, ZZ_Object);
			// Notify the display server about a new object.
			zz_DADD (TheProcess, this);
		}
	}

	// Note: We assume that DisplayActive cannot be set before Root
	// initializes things
};

EObject::~EObject () {

/* ----------------------- */
/* Destructor for EObjects */
/* ----------------------- */

	zz_queue_out (this);
	zz_DREM (this);
	if (zz_nickname != NULL) delete (zz_nickname);
};

long    zz_trunc (long a, int b) {

/* ---------------------------------------------------------- */
/* Truncates  a  to  fit  into  a specified number of decimal */
/* positions                                                  */
/* ---------------------------------------------------------- */

	long    c;

	switch (b) {

		case  1: c = 10; break;
		case  2: c = 100; break;
		case  3: c = 1000; break;
		case  4: c = 10000; break;
		case  5: c = 100000; break;
		case  6: c = 1000000; break;
		case  7: c = 10000000; break;
		case  8: c = 100000000; break;
		case  9: c = 1000000000; break;
		default: return (a);
	}

	if (a >= 0) return (a % c);
	return (- ((-a) % (c / 10)));
}

char    *AI::zz_eid (long eid) {

/* ---------------------------------------------------------- */
/* Returns  textual  event  identifier  for  a  given  AI and */
/* numeric event id                                           */
/* ---------------------------------------------------------- */

	switch (Class) {

		case AIC_timer:

			if (eid == NONE)
				return ("wakeup");
			else
				return (form ("%1d", zz_trunc (eid, 10)));

		case AIC_queue:

			if (eid == ANY)
				return ("any_queue");
			else
				return (form ("%1d", zz_trunc (eid, 10)));

		case AIC_signal:

			if (eid == RECEIVE_b)
				return ("RECEIVE_b");
			else if (eid == RECEIVE_e)
				return ("RECEIVE_e");
			else
				return (form ("%1d", zz_trunc (eid, 10)));

		case AIC_link:

			switch (eid) {

				case ARC_PURGE:
					return ("ARC_PURGE");
				case LNK_PURGE:
					return ("LNK_PURGE");
				case BOT_HEARD:
					return ("BOT_HEARD");
				case EOT_HEARD:
					return ("EOT_HEARD");
				default:
					return ("unknown");
			};

		case AIC_port:

			switch (eid) {

				case SILENCE:
					return ("SILENCE");
				case ACTIVITY:
					return ("ACTIVITY");
				case COLLISION:
					return ("COLLISION");
				case BOJ:
					return ("BOJ");
				case EOJ:
					return ("EOJ");
				case BOT:
					return ("BOT");
				case EOT:
					return ("EOT");
				case BMP:
					return ("BMP");
				case EMP:
					return ("EMP");
				case ANYEVENT:
					return ("ANYEVENT");
				default:
					return ("unknown");
			};

		case AIC_client:
		case AIC_traffic:

			switch (eid) {

				case ARR_MSG:
					return ("ARR_MSG");  // Internal
				case ARR_BST:
					return ("ARR_BST");  // Internal
				case ARRIVAL:
					return ("ARRIVAL");
				case INTERCEPT:
					return ("INTERCEPT");
				default:
					return (form ("arr_Trf%03d",
						zz_trunc (eid, 3)));
			};

		case AIC_process:

			if (eid == DEATH)
				return ("DEATH");
			else if (eid == START)
				return ("START");
			else
				return (((Process*)this)->zz_sn ((int)eid));

#if     OBSERVERS
		case AIC_observer:

			if (eid == OBS_TSERV)
				return ("TIMEOUT");
			else
				return ("unknown");
#endif
		default:
			return ("unknown");
	}
}

/* --------------- */
/* Dummy exposures */
/* --------------- */

sexposure (ZZ_Object)

	excptn ("Object->exposure: Object class cannot be exposed directly");
	USEEMD;
	USESID;
}

sexposure (AI)

	excptn ("AI->exposure: AI class cannot be exposed directly");
	USEEMD;
	USESID;
}
