/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ------------- */
/* The Signal AI */
/* ------------- */

#include        "system.h"

zz_signal::zz_signal () {

/* --------------- */
/* The constructor */
/* --------------- */

	Class = AIC_signal;
	Id = NONE;
}

void    zz_signal::zz_wait (long ev, int pstate) {

/* ----------------------- */
/* The Signal wait request */
/* ----------------------- */

	TIME       t;
	ZZ_REQUEST *rq;
	ZZ_SIG     *s;

	if_from_observer ("Signal->wait: called from an observer");

	if (zz_c_first_wait) {

		zz_c_other = NULL;    // No requests so far

		// Create new simulation event
		zz_c_wait_event = new ZZ_EVENT;
		zz_c_wait_event -> station = TheStation;
		zz_c_wait_event -> process = TheProcess;
	}

#if     ASSERT

	// Check if the requested Signal is not already waited for

	for (rq = TheStation->SWList; rq != NULL; rq = rq->next)
		assert (rq->event_id != ev || rq->event == zz_CE,
			"Signal->wait: duplicate wait for the same Signal");
#endif
	// Check if the Signal has already been received

	for (s = TheStation->RSList; s != NULL; s = s->next) {

		if (s->signum == ev && s->request == NULL) {
			assert (s->RTime <= Time,
		       "Signal->wait: internal error - Signal out of sequence");

			rq = new ZZ_REQUEST (&(TheStation->SWList), this,
				ev, pstate, RQTYPE_SIG,
					s->Info01, s->Info02, (void*) s);
			t = Time;       // Schedule at current time
			s -> request = rq;
			goto SIG_FOUND;
		}
	}

	// Signal not pending: wait for it

	new ZZ_REQUEST (&(TheStation->SWList), this, ev, pstate,
		RQTYPE_SIG, (void*) NONE, (void*) NONE);

	t = TIME_inf;                   // Arrival time unknown

SIG_FOUND:

	assert (zz_c_other != NULL,
		"Signal->wait: internal error -- null request");
	
	if (zz_c_first_wait) {

		zz_c_wait_event -> pstate = pstate;
		zz_c_wait_event -> ai = this;
		zz_c_wait_event -> event_id = ev;
		zz_c_wait_event -> waketime = zz_c_wait_tmin = t;
		zz_c_wait_event -> chain = zz_c_other;
		zz_c_wait_event -> Info01 = zz_c_other -> Info01;
		zz_c_wait_event -> Info02 = zz_c_other -> Info02;

		zz_c_whead = zz_c_other;
		zz_c_other -> when = t;
		zz_c_other -> event = zz_c_wait_event;
		zz_c_other -> other = zz_c_whead;

		zz_c_first_wait = NO;
		if (def (t))
			zz_c_wait_event->enqueue ();
		else
			zz_c_wait_event->store ();
	} else {

		if (zz_c_wait_tmin > t || (zz_c_wait_tmin == t && flip ())) {

			zz_c_wait_event -> pstate = pstate;
			zz_c_wait_event -> ai = this;
			zz_c_wait_event -> event_id = ev;
			zz_c_wait_event -> waketime = zz_c_wait_tmin = t;
			zz_c_wait_event -> chain = zz_c_other;
			zz_c_wait_event -> Info01 = zz_c_other -> Info01;
			zz_c_wait_event -> Info02 = zz_c_other -> Info02;

			if (def (t))
				zz_c_wait_event -> reschedule ();
		}

		zz_c_other -> when = t;
		zz_c_other -> event = zz_c_wait_event;
		zz_c_other -> other = zz_c_whead;
	}
}

int zz_signal::send (long signum, Station *st, void *Info01, void *Info02) {

/* -------------------------------------- */
/* Send a Signal to the specified station */
/* -------------------------------------- */

	ZZ_SIG             *s;
	ZZ_REQUEST         *rq;
	ZZ_EVENT           *e;

	if_from_observer ("Signal->send: called from an observer");

	for (s = st -> RSList; s != NULL;
		s = s->next) {

		if (s->signum == signum) {
			// Overwrite the already pending Signal
			s->Info01 = Info01;
			s->Info02 = Info02;
			s->RTime = Time;
			if (s->request != NULL) {
				s->request->Info01 = Info01;
				s->request->Info02 = Info02;
			}
			return (REJECTED);
		}
	}

	// Add new Signal to the 'received' queue

	s = new ZZ_SIG (signum, Info01, Info02);
	queue_head (s, st->RSList, ZZ_SIG);

	// Check if the station isn't already waiting for the Signal

	for (rq = st -> SWList; rq != NULL; rq = rq -> next) {
		if (rq->event_id == signum) {
			/* Reschedule */
			rq -> when = Time;
			rq -> Info01 = Info01;
			rq -> Info02 = Info02;
			rq -> what = (char*) s;
			s  -> request = rq;

			if ((e = rq -> event) -> waketime > Time || flip ())
					e -> new_top_request (rq);
			return (ACCEPTED);
		}
	}

	return (PENDING);
}

int     zz_signal::sendPS (long signum) {

/* --------------------------------------------- */
/* Send a priority Signal to the current station */
/* --------------------------------------------- */

	ZZ_REQUEST         *rq;

	if_from_observer ("Signal->sendPS: called from an observer");

	// Check if the station is waiting for the Signal

	for (rq = TheStation -> SWList; rq != NULL;
		rq = rq->next) {

		if (rq->event_id == signum) {
			// Reschedule
			rq -> when = Time;
			rq -> Info01 = (void*) signum;
			rq -> Info02 = (void*) (TheStation->Id);

			assert (zz_pe == NULL,
			"Signal->sendPS: more than one pending priority event");

			zz_pe = (zz_pr = rq) -> event;
			return (OK);
		}
	}
	return (ERROR);         // Nobody waiting for the Signal
}

#if     BIG_precision == 1 || ZZ_ATT
inline
#endif
void    ZZ_SIG::cleanup () {

/* -------------------------------- */
/* Clean up after removing a Signal */
/* -------------------------------- */

	ZZ_EVENT           *e;

	if (request == NULL) return;            // Nothing to clean up
	request->when = TIME_inf;               // Withdraw the request
	request->what = NULL;
	Info01 = Info02 = (void*) NONE;
	if ((e = request->event)->chain != request)
		return;                         // No need to reschedule
	e->new_top_request (request->min_request ());
						// New wakeup time
}

int     zz_signal::clear () {

/* ---------------------------------------- */
/* Clear all pending signals at the station */
/* ---------------------------------------- */

	ZZ_SIG  *s;
	int     fg;

	if_from_observer ("Signal->clear: called from an observer");

	for (fg = ERROR; (s = TheStation->RSList) != NULL; ) {
		queue_out (s);
		s->cleanup ();
		delete (s);
		fg = OK;
	}
	return (fg);
}

int     zz_signal::clear (long num) {

/* ---------------------------------------- */
/* Clear Signal number 'num' at the station */
/* ---------------------------------------- */

	ZZ_SIG     *s;

	if_from_observer ("Signal->clear: called from an observer");

	for (s = TheStation->RSList; s != NULL; s = s->next) {

		if (s->signum == num) {
			queue_out (s);
			s->cleanup ();
			delete (s);
			return (OK);
		}
	}
	return (ERROR);
}

sexposure (zz_signal)

/* ----------------------------------- */
/* Defines exposures for the Signal AI */
/* ----------------------------------- */

	sonpaper {

		sexmode (0)

			exPrint0 (Hdr, (int) SId);

		sexmode (1)

			exPrint1 (Hdr, (int) SId);

		sexmode (2)

			exPrint2 (Hdr, (int) SId);
	}

	sonscreen {

		sexmode (0)

			exDisplay0 ((int) SId);

		sexmode (1)

			exDisplay1 ((int) SId);

		sexmode (2)

			exDisplay2 ((int) SId);
	}
}

void    zz_signal::exPrint0 (char *hdr, int sid) {

/* ---------------------------- */
/* Print the full request queue */
/* ---------------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;
	long               l;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Full AI wait list";
		if (isStationId (sid))
			Ouf << " for " << idToStation (sid)->getOName ();
		Ouf << ":\n\n";
	}

	if (!isStationId (sid))
		Ouf << "       Time   St";
	else
		Ouf << "           Time ";

	Ouf << "    Process/Idn     Signal     SState         AI/Idn" <<
		"      State\n";

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (e->chain == NULL || e->station == NULL) continue;
		if (isStationId (sid) && ident (e->station) != sid) continue;

		for (r = e->chain;;) {

			if (r->ai->Class == AIC_signal) {

				if (!isStationId (sid))
					ptime (r->when, 11);
				else
					ptime (r->when, 15);

				if (e->waketime < r->when)
					Ouf << ' ';     // Obsolete
				else if (e->chain == r)
					Ouf << '*';     // Uncertain
				else
					Ouf << '?';

				if (!isStationId (sid)) {
					// Identify the station
					if (e->station != NULL &&
						ident (e->station) >= 0)
						Ouf << form ("%4d ",
					zz_trunc (ident (e->station), 3));
					else
						Ouf << " Sys ";
				}
				print (e->process->getTName (), 10);
				Ouf << form ("/%03d ",
					zz_trunc (e->process->Id, 3));

				print (zz_eid (r->event_id), 10);
				Ouf << ' ';

				print (e->process->zz_sn (r->pstate), 10);
				Ouf << ' ';

				print (e->ai->getTName (), 10);
				if ((l = e->ai->Id) != NONE) {
					if (e->ai->Class == AIC_port)
						l = l & 0177777;
					Ouf << form ("/%03d ",
						zz_trunc (l, 3));
				} else
					Ouf << "     ";

				print (e->process->zz_sn (e->chain->pstate),10);
				Ouf << '\n';
			}

			if ((r = r->other) == e->chain) break;
		}
	}

	Ouf << "\n(" << getOName () << "End of list\n\n";
}

void    zz_signal::exPrint1 (char *hdr, int sid) {

/* ----------------------------------- */
/* Print the abbreviated request queue */
/* ----------------------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Abbreviated AI wait list";
		if (isStationId (sid))
			Ouf << " for " << idToStation (sid) -> getOName ();
		Ouf << ":\n\n";
	}

	if (!isStationId (sid))
		Ouf << "                Time     St";
	else
		Ouf << "                      Time ";
	Ouf << "      Process/Ident     Signal        State\n";

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (e->chain == NULL || e->station == NULL) continue;
		if (isStationId (sid) && ident (e->station) != sid) continue;

		for (r = e->chain;;) {

			if (r->ai->Class == AIC_signal) {

				if (!isStationId (sid))
					ptime (r->when, 20);
				else
					ptime (r->when, 26);

				if (e->waketime < r->when)
					Ouf << ' ';     // Obsolete
				else if (e->chain == r)
					Ouf << '*';     // Uncertain
				else
					Ouf << '?';

				if (!isStationId (sid)) {
					// Identify the station
					if (e->station != NULL &&
						ident (e->station) >= 0)
						Ouf << form ("%6d ",
					zz_trunc (ident (e->station), 5));
					else
						Ouf << "   Sys ";
				}
				print (e->process->getTName (), 12);
				Ouf << form ("/%05d ",
					zz_trunc (e->process->Id, 5));

				print (zz_eid (e->event_id), 10);
				Ouf << ' ';

				print (e->process->zz_sn (r->pstate), 12);
				Ouf << '\n';
			}

			if ((r = r->other) == e->chain) break;
		}
	}

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void    zz_signal::exPrint2 (char *hdr, int sid) {

/* ------------------------------- */
/* Print the Signal queue contents */
/* ------------------------------- */

	int             l, i;
	ZZ_SIG          *s, **ss;
	TIME            t;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Pending items";
		if (isStationId (sid))
			Ouf << " for " << idToStation (sid)->getOName ();
		Ouf << ":\n\n";
	}

	Ouf << "                Time      Signal";
	if (!isStationId (sid)) Ouf << " Station";
	Ouf << '\n';

	if (isStationId (sid)) {

		for (s = idToStation (sid)->RSList; s != NULL; s = s->next) {

			print (s->RTime, 20);
			Ouf << "  ";
			print (Signal->zz_eid (s->signum), 10);
			Ouf << '\n';
		}
	} else {

		ss = new ZZ_SIG* [NStations];
		for (i = 0; i < NStations; i++)
			ss [i] = idToStation (i)->RSList;

		while (1) {
			for (t = TIME_inf, s = NULL, i = 0; i < NStations; i++)
				if (ss [i] != NULL && ss [i]->RTime < t)
					t = (s = ss [l = i]) -> RTime;

			if (s == NULL) break;

			print (t, 20);
			Ouf << "  ";
			print (Signal->zz_eid (s->signum), 10);
			Ouf << ' ';
			print (i, 7);
			Ouf << '\n';
		}

		delete (ss);
	}

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void    zz_signal::exDisplay0 (int sid) {

/* ------------------------------ */
/* Display the full request queue */
/* ------------------------------ */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;
	long               l;

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (e->chain == NULL || e->station == NULL) continue;
		if (isStationId (sid) && ident (e->station) != sid) continue;

		for (r = e->chain;;) {

			if (r->ai->Class == AIC_signal) {

				dtime (r->when);
				if (e->waketime < r->when)
					display (' ');
				else if (e->chain == r)
					display ('*');
				else
					display ('?');

				if (!isStationId (sid)) {
					// Identify the station
					if (e->station != NULL &&
						ident (e->station) >= 0)
						   display (ident (e->station));
					else
						   display ("Sys");
				}
				display (e->process->getTName ());
				display (e->process->Id);
				display (zz_eid (r->event_id));
				display (e->process->zz_sn (r->pstate));
				display (e->ai->getTName ());
				if ((l = e->ai->Id) != NONE) {
					if (e->ai->Class == AIC_port)
						l = l & 0177777;
					display (l);
				} else
					display (' ');

				display (e->process->zz_sn (e->chain->pstate));
			}
			if ((r = r->other) == e->chain) break;
		}
	}
}

void    zz_signal::exDisplay1 (int sid) {

/* ------------------------------------- */
/* Display the abbreviated request queue */
/* ------------------------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (e->chain == NULL || e->station == NULL) continue;
		if (isStationId (sid) && ident (e->station) != sid) continue;

		for (r = e->chain;;) {

			if (r->ai->Class == AIC_signal) {

				dtime (r->when);
				if (e->waketime < r->when)
					display (' ');
				else if (e->chain == r)
					display ('*');
				else
					display ('?');

				if (!isStationId (sid)) {
					// Identify the station
					if (e->station != NULL &&
						ident (e->station) >= 0)
						   display (ident (e->station));
					else
						   display ("Sys");
				}
				display (e->process->getTName ());
				display (e->process->Id);
				display (zz_eid (e->event_id));
				display (e->process->zz_sn (r->pstate));
			}
			if ((r = r->other) == e->chain) break;
		}
	}
}

void    zz_signal::exDisplay2 (int sid) {

/* --------------------------------- */
/* Display the Signal queue contents */
/* --------------------------------- */

	int             l, i;
	ZZ_SIG          *s, **ss;
	TIME            t;

	if (isStationId (sid)) {

		for (s = idToStation (sid)->RSList; s != NULL; s = s->next) {

			display (s->RTime);
			display (Signal->zz_eid (s->signum));
		}
	} else {
		ss = new ZZ_SIG* [NStations];
		for (i = 0; i < NStations; i++)
			ss [i] = idToStation (i)->RSList;

		while (1) {
			for (t = TIME_inf, s = NULL, i = 0; i < NStations; i++)
				if (ss [i] != NULL && ss [i]->RTime < t)
					t = (s = ss [l = i]) -> RTime;

			if (s == NULL) break;

			display (t);
			display (Signal->zz_eid (s->signum));
			display (i);
		}

		delete (ss);
	}
}
