/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */


/* ----------------------- */
/* Make program for smurph */
/* ----------------------- */

#include        "version.h"

#if	ZZ_OWNLIB == 1
#include	"LIB/lib.h"
extern "C" {
int	system (const char*);
int	unlink (const char*);
#ifndef	ZZ_ATT
int	mkdir (const char*, int);
int	chdir (const char*);
#endif
char	*getwd (char*);	// Seems to be missing
}
long 		MemoryUsed;
#else
#include        <stream.h>
#endif

#if	ZZ_ATT
#include	<osfcn.h>
#endif

#include        <sys/types.h>
#include        <sys/param.h>
#include	<sys/fcntl.h>
extern  "C" {
#include        <sys/dir.h>
int             flock (int, int);
}
#define	KERNEL	// Bypassing a bug in stat.h on some machines
#include        <sys/stat.h>
#undef	KERNEL
#include        <signal.h>
#include	<ctype.h>

#ifdef	sgi
typedef	void (*SIGARG) (int, ...);
#else
#ifdef	ZZ_ATT_SUN
#define	SIGARG	SIG_PF
#else
typedef	void (*SIGARG) ();
#endif
#endif

/* ---------------------------------------------------- */
/* The three constants below should be defined by maker */
/* ---------------------------------------------------- */
#ifndef CCOMP
#define CCOMP   "g++"
#endif

#ifndef MAXLIB
#define MAXLIB  5
#endif

#ifndef	ZZ_INCPATH
#define	ZZ_INCPATH	""
#endif

	//
	// Call format:
	//
	//  mks params filenames
	//
	//  where params may contain the following items
	//
	//      -f           -- optimization (default = no optimization)
	//      -a           -- assertions switched off (default = on)
	//      -v           -- observers switched off (default = on)
	//      -m           -- suppress error detection for MPA (default = on)
	//      -d           -- DISTANCE should be BIG (default = off)
	//      -i           -- BITCOUNT should be BIG (default = off)
	//      -r           -- RATE should be BIG (default = off)
	//      -n           -- clock errors switched off (default = on)
	//	-q	     -- message queue size can be limited
	//      -g           -- debugging (default = off)
	//      -u           -- standard client disabled (default = enabled)
	//      -b n         -- TIME precision (default = 2)
	//      -o fname     -- send binaries to 'fname' (default 'smurph')
	//      -t           -- 'touch' the protocol source files
	//
	//		The following options are 'undocumented'. If you have
	//		managed to get this far, you can as well learn about
	//		them:
	//
	//	-E           -- leave the '.C' files after compilation
	//	-V	     -- 'verbose'
	//
	// The configuration of options, with exception of '-o', determines the
	// library subdirectory name. The default configuration of options
	// corresponds to name:  nyyynnnyny1
	//
	// filenames describe the files containing the smurph code. Suffixes:
	// '.c' and '.C' are legal and treated in the same way. Suffix '.C' is
	// reserved for the output file produced by smpp. If no files
	// are specified, all files in the current directory ending with the
	// proper suffixes are assumed.
	//

char    PRC [24], OPT [16], ASR [10], OBS [10], AER [10], DST [10], BTC [10],
	RTS [10], TOL [10], QSL [10], R48 [10], DBG [16], CLI [10];

char    *options [24],
	*ofname, libfname [16], **cfiles, *pname;

int     ncfiles = 0, cfsize = 16;

char    cd [MAXPATHLEN];

char    tmpfname [48];

#define LOCK_SH 1               // Cannot include sys/file.h
#define LOCK_EX 2
#define LOCK_UN 8

int	FError,			// File error handler
	LeaveCFiles = 0,	// Flag == do not remove preprocessor '.C' files
	BeVerbose = 0,		// Inform about what exactly is being done
	TouchSources = 0;	// Flag == touch source files

void    badArgs () {

	cerr << "Usage:  " << pname << " 'options' 'files'\n";
	cerr << "      Options:\n";
	cerr << "       -f         optimization\n";
	cerr << "       -a         assertions switched off\n";
	cerr << "       -v         observers deactivated\n";
	cerr << "       -m         error detection for BIG arith. suppressed\n";
	cerr << "       -d         distances are BIG\n";
	cerr << "       -i         bit counters are BIG\n";
	cerr << "       -r         port transmission rates are BIG\n";
	cerr << "       -n         clocks are absolutely accurate\n";
	cerr << "       -q         message queue size can be limited\n";
	cerr << "       -8         use the standard drand48 rnd gen family\n";
	cerr << "       -g         debugging\n";
	cerr << "       -u         standard client permanently disabled\n";
	cerr << "       -b n       precision of BIG numbers (default is 2)\n";
	cerr << "       -o fname   direct binary simulator to 'fname'\n";
	cerr << "       -t         'touch' the protocol source files\n";
	cerr << "      Default: '-b 2 -o smurph' (other flags cleared)\n";
	exit (1);
}

void    sigint () {

	unlink (tmpfname);
	exit (1);
}

void	excptn (char *t) {

	cerr << pname << ": " << t << '\n';
	sigint ();
}

int     processFName (char *fn) {

/* --------------------------- */
/* Processes filename argument */
/* --------------------------- */

	char    *cp, **scratch;
	int     i;

	for (cp = fn + strlen (fn) - 1; cp != fn; cp--)
		// Locate the suffix
		if (*cp == '.') break;

	if (cp == fn) return (-1);

	cp++;

	if (strcmp (cp, "c") && strcmp (cp, "cc"))
		// Illegal suffix
		return (-1);
#if	ZZ_ATT
	// Get rid of the ..c crap
	if (*(cp-2) == '.') return (-1);
#endif

	// Check if the file has not been specified already

	for (i = 0; i < ncfiles; i++)
		if (strcmp (cfiles [i], fn) == 0) return (-1);

	// Add the new file to the list

	if (ncfiles >= cfsize) {
		// Must increase the array size
		scratch = new char* [cfsize];
		for (i = 0; i < cfsize; i++) scratch [i] = cfiles [i];
		delete (cfiles);
		cfiles = new char* [cfsize += cfsize];
		for (i = 0; i < ncfiles; i++) cfiles [i] = scratch [i];
		delete (scratch);
	}

	cfiles [ncfiles] = new char [strlen (fn) + 1];
	strcpy (cfiles [ncfiles], fn);
	ncfiles++;
	return (0);
}

char    *toObject (char *fn) {

/* --------------------------------------------- */
/* Converts source file name to object file name */
/* --------------------------------------------- */

	char    *cp;
	int     i;

	for (cp = fn + strlen (fn) - 1; cp != fn; cp--)
		// Locate the suffix
		if (*cp == '.') break;

	for (i = 0; ; i++, fn++) {
		cd [i] = *fn;
		if (fn == cp) break;
	}

	cd [++i] = 'o';
	cd [++i] = '\0';
	return (cd);
}

char    *toSmpp (char *fn) {

/* ---------------------------------------------------------- */
/* Converts source file name to preprocessor output file name */
/* ---------------------------------------------------------- */

	char    *cp;
	int     i;

	for (cp = fn + strlen (fn) - 1; cp != fn; cp--)
		// Locate the suffix
		if (*cp == '.') break;

	for (i = 0; ; i++, fn++) {
		cd [i] = *fn;
		if (fn == cp) break;
	}

	cd [++i] = 'C';
	cd [++i] = '\0';
	return (cd);
}

void    getFNames () {

/* ---------------------------------------------------------- */
/* Searches  the  current  directory  for  filenames with the */
/* proper suffixes                                            */
/* ---------------------------------------------------------- */

	DIR     *cdir;
	struct  direct  *dentry;

	if ((cdir = opendir (".")) == NULL)
		excptn ("cannot open current directory");

	while ((dentry = readdir (cdir)) != NULL) {
		// Process next entry
		processFName (dentry->d_name);
	}

	closedir (cdir);
}

int     findLVer () {

/* ---------------------------------------------------------- */
/* Finds  and  possibly  creates  the  proper library version */
/* subdirectory                                               */
/* ---------------------------------------------------------- */

	DIR     *cdir;
	struct  direct  *dentry;
	struct  stat    statbuf;
	int     count, tfd;
	long    matime, loc, cloc;
	char    *cc;

	if ((cdir = opendir (".")) == NULL)
		excptn ("cannot open library directory");

	for (count = 0; (dentry = readdir (cdir)) != NULL; ) {
		if (*(dentry->d_name) == '.') continue;
		if (strcmp (dentry->d_name, libfname) == 0) break;
		count++;
	}

	if (dentry != NULL) {
		// Directory found
		closedir (cdir);
		cc = form ("%s/lockfile", libfname);
		if ((tfd = open (cc, O_WRONLY, 0)) < 0)
			// Is supposed to exist
			tfd = open (cc, O_WRONLY+O_CREAT, 0777);
		return (tfd);
	}

	// Check if something should be removed

	while (count >= MAXLIB) {

		// Determine the least recently used directory
		rewinddir (cdir);
		for (matime = -1, cloc = telldir (cdir);
			(dentry = readdir (cdir)) != NULL;
				cloc = telldir (cdir)) {

			if (*(dentry->d_name) == '.') continue;
			cc = form ("%s/main.o", dentry->d_name);
			// main.o should always be present there
			if ((tfd = open (cc, O_RDONLY, 0)) < 0) {
				// Garbage -- feel free to remove it
				tfd = open (form ("%s/lockfile",dentry->d_name),
					O_WRONLY, 0);
				flock (tfd, LOCK_EX);
				cout << "Removing: " << dentry->d_name << '\n';
				system (form ("rm -fr %s 2>/dev/null",
							      dentry->d_name));
				flock (tfd, LOCK_UN);
				close (tfd);
				goto NEXT;
			}

			// Determine the status of main.o
			if (fstat (tfd, &statbuf) < 0)
				excptn (form ("problems accessing '%s'", cc));

			close (tfd);

			if (matime == -1 || statbuf.st_atime < matime) {
				matime = statbuf.st_atime;
				loc = cloc;
			}
		}

		if (matime == -1) {
PBMS:
			excptn ("problems determining victim library");
		}

		seekdir (cdir, loc);
		if ((dentry = readdir (cdir)) == NULL) goto PBMS;
		tfd = open (form ("%s/lockfile", dentry->d_name), O_WRONLY, 0);
		flock (tfd, LOCK_EX);
		cout << "Removing: " << dentry->d_name << '\n';
		system (form ("rm -fr %s 2> /dev/null", dentry->d_name));
		flock (tfd, LOCK_UN);
		close (tfd);

NEXT: 		count--;
	}
	closedir (cdir);

	// Now create the new library

	if (mkdir (libfname, 0777) < 0)
		excptn (form ("cannot create '%s'", libfname));

	return (open (form ("%s/lockfile", libfname), O_WRONLY+O_CREAT, 0777));
}

void    makeSmurph () {

/* ---------------------------- */
/* Makes the simulator instance */
/* ---------------------------- */

	char    fname [MAXPATHLEN], c, *cp;
	int     i, eraseo;

	istream *inp;
	ostream *out;

	// Check if the .ssignature file is present

	eraseo = 1;
	FError = 0;
	inp = new istream (".ssignature", "r");
	if (!FError) {
		for (i = 0; i < MAXPATHLEN-1; ) {
			if (inp->eof ()) break;
			inp->get (c);
			if (isspace (c)) continue;
			fname [i++] = c;
		}

		eraseo = strcmp (fname, libfname);
		inp->close ();
	}

	if (eraseo || TouchSources) {

		// Erase old binaries
		for (i = 0; i < ncfiles; i++)
			system (form ("rm -fr %s", toObject (cfiles [i])));

		FError = 0;
		out = new ostream (".ssignature", "w");
		if (FError)
			excptn ("cannot open .ssignature file");

		*out << libfname << '\n';
		out->close ();
	}

	strcpy (fname, ZZ_SOURCES);
	strcat (fname, "/");
	strcat (fname, "SIMULATOR/Dependencies");

	FError = 0;
	inp = new istream (fname, "r");
	if (FError)
		excptn ("cannot open 'Dependencies' file");

	// Now create tmp file
	strcpy (tmpfname, "/tmp/mksmurphXXXXXX");
	mktemp (tmpfname);

	if (signal (SIGINT, SIG_IGN) != SIG_IGN)
		signal (SIGINT, (SIGARG) sigint);
	if (signal (SIGQUIT, SIG_IGN) != SIG_IGN)
		signal (SIGQUIT, (SIGARG) sigint);

	FError = 0;
	out = new ostream (tmpfname, "w");
	if (FError)
		excptn ("cannot open temporary file");

	*out << "SRC= " << ZZ_SOURCES << '\n';
	*out << "LIB= " << ZZ_LIBPATH << '/' << libfname << '\n';
#ifdef	ZZ_INCPATH
	*out << "INC= " << ZZ_INCPATH << '\n';
#endif

	*out << "CMP= " << CCOMP;
#if	ZZ_ATT
	*out << " -w";	// Shut up ccomp complaining about not reached
			// while (0)'s preceded by return
#endif
	if (BeVerbose) *out << " -v";
	*out << '\n';
	*out << "OPT= " << OPT << '\n';
	*out << "ASR= " << ASR << '\n';
	*out << "OBS= " << OBS << '\n';
	*out << "AER= " << AER << '\n';
	*out << "DST= " << DST << '\n';
	*out << "BTC= " << BTC << '\n';
	*out << "RTS= " << RTS << '\n';
	*out << "TOL= " << TOL << '\n';
	*out << "QSL= " << QSL << '\n';
	*out << "R48= " << R48 << '\n';
	*out << "DBG= " << DBG << '\n';
	*out << "CLI= " << CLI << '\n';
	*out << "PRC= " << PRC << '\n';
#if ZZ_OWNLIB == 1
	*out << "OLP= " << ZZ_SOURCES << "/LIB/lib.o " << ZZ_SOURCES <<
		"/LIB/iolink.o"
#ifdef	mips
		<< ' ' << ZZ_SOURCES << "/LIB/mips.o"
#endif
							<< '\n';
	*out << "OLB= -DZZ_OLB\n";
#else
	*out << "OLP=\n";
	*out << "OLB=\n";
#endif
	*out << "TRG= " << ofname << '\n';

	// The list of user object files

	*out << "USO=";
	for (i = 0; i < ncfiles; i++)
		*out << ' ' << toObject (cfiles [i]);
	*out << '\n';

	// Copy the rest

	while (! inp->eof ()) {

		inp->get (c);

		if (c == '?') {
			// Pattern
			if (libfname [10] != 'y') {
				// Strip the executable unless
				// -g has been selected
				*out << "\tstrip " << ofname << '\n';
			}
			for (i = 0; ; i++) {
				if (inp->eof () || i >= MAXPATHLEN-1)
			    excptn ("'Dependencies' file has illegal contents");
				inp->get (c);
				if (c == '?') break;
				fname [i] = c;
			}

			fname [i] = '\0';

			for (i = 0; i < ncfiles; i++) {
				for (cp = fname; *cp != NULL; cp++) {
					if (*cp == '%')
						*out << cfiles [i];
					else if (*cp == '@')
						*out << toObject (cfiles [i]);
					else if (*cp == '#')
						*out << i;
					else if (*cp == '&')
						*out << toSmpp (cfiles [i]);
					else if (*cp == '!')
						*out << i;
					else
						out->put (*cp);
				}
				if (i < ncfiles-1) *out << '\n';
			}
		} else {
		
			out->put (c);
		}
	}

	inp->close ();
	out->close ();

	eraseo = BeVerbose ? system (form ("make -f %s", tmpfname)) :
				system (form ("make -s -f %s", tmpfname));

	unlink (tmpfname);
	// Remove any leftover '*.C' files
	if (!LeaveCFiles)
		for (i = 0; i < ncfiles; i++) unlink (toSmpp (cfiles [i]));

	if (eraseo)
		excptn ("errors -- simulator not created");

	if (signal (SIGINT, SIG_IGN) != SIG_IGN) signal (SIGINT, SIG_DFL);
	if (signal (SIGQUIT, SIG_IGN) != SIG_IGN) signal (SIGQUIT, SIG_DFL);
}

void	fError (char *t) {

// File error handler

#if	ZZ_ATT
	t++;	// To avoid complaints
#endif
	FError = 1;
}

main    (int argc, char *argv[]) {

	int     nopt, i, semfd, verfd;

	// Initialize things for argument processing

	set_File_error_handler (fError);

	ofname = NULL;

	pname = *argv;
	argc--; argv++;

	strcpy (OPT, "-DZZ_OPT=0");             options [0 ] = OPT;
	strcpy (ASR, "-DZZ_ASR=1");             options [1 ] = ASR;
	strcpy (OBS, "-DZZ_OBS=1");             options [2 ] = OBS;
	strcpy (AER, "-DZZ_AER=1");             options [3 ] = AER;
	strcpy (DST, "-DZZ_DST=0");             options [4 ] = DST;
	strcpy (BTC, "-DZZ_BTC=0");             options [5 ] = BTC;
	strcpy (RTS, "-DZZ_RTS=0");             options [6 ] = RTS;
	strcpy (TOL, "-DZZ_TOL=1");             options [7 ] = TOL;
	strcpy (QSL, "-DZZ_QSL=0");             options [8 ] = QSL;
	strcpy (R48, "-DZZ_R48=0");		options [9 ] = R48;
	strcpy (DBG, "-DZZ_DBG=0");             options [10] = DBG;
	strcpy (CLI, "-DZZ_CLI=1");             options [11] = CLI;
	strcpy (PRC, "-DBIG_precision=2");      options [12] = PRC;
						options [13] = NULL;

	nopt = 13;

	strcpy (libfname, "xxxxxxxxxxxxxxx");

	cfiles = new char* [16];

	while (argc > 0) {

		// Process next argument

		if (**argv != '-' || (*argv)[1] == '\0' || (*argv)[2] != '\0') {
			if (processFName (*argv)) badArgs ();
			argv++;
			argc--;
			continue;
		}

		switch ((*argv)[1]) {

		  case 'f' :

			if (libfname [0] != 'x') badArgs ();
			strcpy (&(OPT [9]), "1 -O");
			libfname [0] = 'y';
			break;

		  case 'a' :

			if (libfname [1] != 'x') badArgs ();
			ASR [9] = '0';
			libfname [1] = 'n';
			break;

		  case 'v' :

			if (libfname [2] != 'x') badArgs ();
			OBS [9] = '0';
			libfname [2] = 'n';
			break;

		  case 'm' :

			if (libfname [3] != 'x') badArgs ();
			AER [9] = '0';
			libfname [3] = 'n';
			break;

		  case 'd' :

			if (libfname [4] != 'x') badArgs ();
			DST [9] = '1';
			libfname [4] = 'y';
			break;

		  case 'i' :

			if (libfname [5] != 'x') badArgs ();
			BTC [9] = '1';
			libfname [5] = 'y';
			break;

		  case 'r' :

			if (libfname [6] != 'x') badArgs ();
			RTS [9] = '1';
			libfname [6] = 'y';
			break;

		  case 'n' :

			if (libfname [7] != 'x') badArgs ();
			TOL [9] = '0';
			libfname [7] = 'n';
			break;

		  case 'q' :

			if (libfname [8] != 'x') badArgs ();
			QSL [9] = '1';
			libfname [8] = 'y';
			break;

		  case '8' :

			if (libfname [9] != 'x') badArgs ();
			R48 [9] = '1';
			libfname [9] = 'y';
			break;

		  case 'g' :

			if (libfname [10] != 'x') badArgs ();
			strcpy (&(DBG [9]), "1 -g");
			libfname [10] = 'y';
			break;

		  case 'u' :

			if (libfname [11] != 'x') badArgs ();
			CLI [9] = '0';
			libfname [11] = 'n';
			break;

		  case 'b' :

			if (argc < 2 || libfname [11] != 'x') badArgs ();
			argv++;
			argc--;

			i = **argv - '0';
			if (i < 0 || i > 9 || (*argv)[1] != '\0') badArgs ();
			libfname [12] = **argv;
			if (**argv == '0') {
				// Convert to TYPE_double (-1)
				PRC [16] = '\0';
				strcat (PRC, "-1");
			} else {
				PRC [16] = **argv;
			}
			break;

		  case 'o' :

			if (argc < 2 || ofname != NULL) badArgs ();
			argv++;
			argc--;

			ofname = new char [strlen (*argv) + 1];
			strcpy (ofname, *argv);
			break;

 		  case 't' :

			if (TouchSources++) badArgs ();
			break;

		  case 'E' :

			LeaveCFiles = 1;
			break;

		  case 'V' :

			BeVerbose = 1;
			break;

		  default :

			if (processFName (*argv)) badArgs ();;
		}

		argv++;
		argc--;
	}

	if (ofname == NULL) {
		ofname = new char [16];
		strcpy (ofname, "smurph");
	}

	// Prepare the library name

	for (i = 0; i < nopt-1; i++) {

		if ((options [i])[9] == '1')
			libfname [i] = 'y';
		else
			libfname [i] = 'n';

	}

	// Now the precision of time

	libfname [nopt-1] = PRC [16];
	libfname [nopt] = '\0';

	if (ncfiles == 0) getFNames ();

	if (ncfiles == 0) excptn ("no files to compile");

	getwd (cd);     // Save the current directory

	if (chdir (ZZ_LIBPATH) < 0)
		excptn (form ("cannot access '%s'", ZZ_LIBPATH));

	// Try to open tentatively the library directory
	if ((semfd = open (".", O_RDONLY, 0)) < 0)
		excptn (form ("cannot open '%s'", ZZ_LIBPATH));
	close (semfd);

	// Now open the version file to be used as a semaphore
	if ((semfd = open (form ("%s/version.h", ZZ_SOURCES), O_WRONLY, 0)) < 0)
		excptn ("cannot open version.h");

	// Hard lock: make sure nothing changes in the library for a while

	if (flock (semfd, LOCK_EX) < 0)
		cerr << pname << ": warning -- cannot lock library access\n";

	// Check if the library exists

	if ((verfd = findLVer ()) < 0)
		excptn ("cannot open library version subdirectory");

	// Lock the library to make sure that another copy of mks doesn't
	// attempt to modify/remove it

	if (flock (verfd, LOCK_EX) < 0)
		cerr << pname << ": warning -- cannot lock '" << ZZ_LIBPATH <<
			'/' << libfname << "'\n";

	flock (semfd, LOCK_UN);         // Others can proceed (as long as they
	close (semfd);                  // don't try to remove our version)

	// That's it. Now prepare the makefile for the library

	chdir (cd);                     // Resume original directory
	makeSmurph ();
}
